# api.py

from typing import Dict, Any
import os

def rename_file(old_file_path: str, new_file_path: str) -> Dict[str, Any]:
    """
    Renames a file from old_file_path to new_file_path.

    Args:
        old_file_path (str): The current path of the file.
        new_file_path (str): The new path for the file.

    Returns:
        Dict[str, Any]: A dictionary containing 'output', 'error', and 'error_code'.
    """
    try:
        if not old_file_path.strip() or not new_file_path.strip():
            return {
                "output": None,
                "error": "Both old_file_path and new_file_path must be provided.",
                "error_code": "MISSING_PATHS"
            }

        resolved_old_path = os.path.abspath(os.path.expanduser(old_file_path))
        resolved_new_path = os.path.abspath(os.path.expanduser(new_file_path))
        new_dir = os.path.dirname(resolved_new_path)

        if not os.path.exists(resolved_old_path):
            return {
                "output": None,
                "error": f"File not found: {resolved_old_path}",
                "error_code": "FILE_NOT_FOUND"
            }

        if os.path.exists(resolved_new_path):
            return {
                "output": None,
                "error": f"Destination already exists: {resolved_new_path}",
                "error_code": "DESTINATION_EXISTS"
            }

        # Ensure the destination directory exists
        if not os.path.exists(new_dir):
            try:
                os.makedirs(new_dir, exist_ok=True)
            except PermissionError as e:
                return {
                    "output": None,
                    "error": f"Permission denied while creating directories: {e}",
                    "error_code": "PERMISSION_DENIED"
                }
            except OSError as e:
                return {
                    "output": None,
                    "error": f"OS error while creating directories: {e}",
                    "error_code": "OS_ERROR"
                }

        os.rename(resolved_old_path, resolved_new_path)

        return {
            "output": f"File renamed from {resolved_old_path} to {resolved_new_path}",
            "error": None,
            "error_code": "SUCCESS"
        }

    except PermissionError as e:
        return {
            "output": None,
            "error": f"Permission denied: {e}",
            "error_code": "PERMISSION_DENIED"
        }
    except FileExistsError as e:
        return {
            "output": None,
            "error": f"Destination already exists: {e}",
            "error_code": "DESTINATION_EXISTS"
        }
    except FileNotFoundError as e:
        return {
            "output": None,
            "error": f"File not found: {e}",
            "error_code": "FILE_NOT_FOUND"
        }
    except OSError as e:
        return {
            "output": None,
            "error": f"OS error occurred: {e}",
            "error_code": "OS_ERROR"
        }
    except Exception as e:
        return {
            "output": None,
            "error": f"Unexpected error: {e}",
            "error_code": "UNEXPECTED_ERROR"
        }


# toolkit/file_toolkit/extract_file_as_markdown/api.py

from typing import Any, Dict, Optional, List
import pymupdf4llm  # Ensure this package is installed
import os


def pdf2markdown(
    input_file_path: str,
    pages: Optional[List[int]] = None,
    output_file_path: Optional[str] = None
) -> Dict[str, Any]:
    """
    Extracts content from a PDF file and converts it to Markdown format.
    Optionally, specific pages can be selected, and the output can be saved to a file.

    Args:
        input_file_path (str): The path to the input PDF file.
        pages (List[int], optional): A list of 0-based page numbers to extract. Defaults to None, which extracts all pages.
        output_file_path (str, optional): The path to save the output Markdown file. If not provided, the Markdown text is returned.

    Returns:
        Dict[str, Any]: A dictionary containing 'output', 'error', and 'error_code'.
                        On success, 'output' includes the 'markdown' content or a success message if saved to a file.
    """
    try:
        # Check if the input file exists
        if not os.path.isfile(input_file_path):
            return {
                "output": None,
                "error": f"Input file '{input_file_path}' does not exist.",
                "error_code": "FILE_NOT_FOUND"
            }

        # Convert PDF to Markdown
        try:
            markdown_text = pymupdf4llm.to_markdown(input_file_path, pages)
        except Exception as e:
            return {
                "output": None,
                "error": f"JS_EXECUTION_ERROR: {e}",
                "error_code": "JS_EXECUTION_ERROR"
            }

        if output_file_path:
            # Save the Markdown text to the specified file
            try:
                with open(output_file_path, 'w', encoding='utf-8') as md_file:
                    md_file.write(markdown_text)
                return {
                    "output": {"message": f"Markdown file saved to '{output_file_path}'."},
                    "error": None,
                    "error_code": "SUCCESS"
                }
            except IOError as io_err:
                return {
                    "output": None,
                    "error": f"IO_ERROR: {io_err}",
                    "error_code": "IO_ERROR"
                }
        else:
            # Return the Markdown text
            return {
                "output": {"markdown": markdown_text},
                "error": None,
                "error_code": "SUCCESS"
            }

    except Exception as e:
        return {
            "output": None,
            "error": f"UNEXPECTED_ERROR: {e}",
            "error_code": "UNEXPECTED_ERROR"
        }


# tools/search_file_or_folder/test.py

import pytest
from unittest.mock import patch
from .api import search_file_or_folder

@pytest.mark.asyncio
async def test_search_file_or_folder_success_both():
    name = "test_file"
    search_path = "/test/path"
    search_type = "both"
    case_sensitive = True
    partial_match = False
    expected_matched_paths = ["/test/path/test_file.txt", "/test/path/test_file_folder"]

    with patch('os.walk') as mock_walk, \
         patch('unicodedata.normalize', side_effect=lambda form, s: s.casefold() if not case_sensitive else s):
        
        mock_walk.return_value = [
            ("/test/path", ["test_file_folder"], ["test_file.txt"])
        ]
        
        result = search_file_or_folder(name, search_path, search_type, case_sensitive, partial_match)
        assert result['output']['matched_paths'] == expected_matched_paths
        assert result['error'] is None
        assert result['error_code'] == "SUCCESS"

@pytest.mark.asyncio
async def test_search_file_or_folder_invalid_search_type():
    name = "test_file"
    search_path = "/test/path"
    search_type = "invalid_type"
    case_sensitive = True
    partial_match = False
    expected_error = f"Invalid search_type: {search_type}. Choose from 'file', 'folder', or 'both'."
    
    result = search_file_or_folder(name, search_path, search_type, case_sensitive, partial_match)
    assert result['output'] is None
    assert result['error'] == expected_error
    assert result['error_code'] == "INVALID_SEARCH_TYPE"

@pytest.mark.asyncio
async def test_search_file_or_folder_case_insensitive_partial_match():
    name = "test"
    search_path = "/test/path"
    search_type = "file"
    case_sensitive = False
    partial_match = True
    expected_matched_paths = ["/test/path/TestFile.txt"]

    with patch('os.walk') as mock_walk, \
         patch('unicodedata.normalize', side_effect=lambda form, s: s.casefold()):
        
        mock_walk.return_value = [
            ("/test/path", [], ["TestFile.txt", "AnotherFile.doc"])
        ]
        
        result = search_file_or_folder(name, search_path, search_type, case_sensitive, partial_match)
        assert result['output']['matched_paths'] == expected_matched_paths
        assert result['error'] is None
        assert result['error_code'] == "SUCCESS"

@pytest.mark.asyncio
async def test_search_file_or_folder_no_matches():
    name = "nonexistent"
    search_path = "/test/path"
    search_type = "file"
    case_sensitive = True
    partial_match = False
    expected_matched_paths = []

    with patch('os.walk') as mock_walk, \
         patch('unicodedata.normalize', side_effect=lambda form, s: s):
        
        mock_walk.return_value = [
            ("/test/path", [], ["file1.txt", "file2.doc"])
        ]
        
        result = search_file_or_folder(name, search_path, search_type, case_sensitive, partial_match)
        assert result['output']['matched_paths'] == expected_matched_paths
        assert result['error'] is None
        assert result['error_code'] == "SUCCESS"

@pytest.mark.asyncio
async def test_search_file_or_folder_exception():
    name = "test_file"
    search_path = "/test/path"
    search_type = "file"
    case_sensitive = True
    partial_match = False

    with patch('os.walk', side_effect=Exception("Walk error")), \
         patch('unicodedata.normalize'):
        
        result = search_file_or_folder(name, search_path, search_type, case_sensitive, partial_match)
        assert result['output'] is None
        assert result['error'] == "Error searching for file or folder: Walk error"
        assert result['error_code'] == "SEARCH_FAILED"


from typing import Callable, Dict, List, Optional, Any


class OperatorNode:
    def __init__(self, name: str, func: Callable, input_links: Optional[List[str]] = None, condition=None):
        self.name = name
        self.func = func
        self.input_links = input_links if input_links is not None else []
        self.output_links = []
        self.status = 'idle'
        self.outputs = None
        self.condition = condition

    def execute(self, inputs: dict):
        if self.status != 'idle':
            raise RuntimeError(f"Operator {self.name} is not in 'idle' state. Current state: {self.status}")

        if self.condition is None or self.condition():
            print(f"Running operator {self.name}")
            self.status = "running"
            self.outputs = self.func(inputs)
            self.status = 'completed'
            return self.outputs
        else:
            print(f"Skipping execution of {self.name} due to unmet condition.")
            self.status = "skipped"
            return None

    def __call__(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        return self.execute(inputs)


class ComputationGraph:
    def __init__(self):
        self.operator_nodes = {}
        self.execution_order = None
        self.context = {}  

    def add_operator(self, operator_node):
        self.operator_nodes[operator_node.name] = operator_node
        self.execution_order = None  # Reset the execution order

    def build(self):
        visited = set()
        self.execution_order = []

        def visit(op_name):
            if op_name not in visited:
                visited.add(op_name)
                if op_name in self.operator_nodes:
                    for dep in self.operator_nodes[op_name].input_links:
                        visit(dep)
                    self.execution_order.append(op_name)

        for op_name in self.operator_nodes:
            visit(op_name)
    
    def reset(self):
        for operator in self.operator_nodes.values():
            operator.status = 'idle'
            operator.outputs = None
        self.execution_order = None
        self.context = {}
 
    def run(self, inputs):
        self.context = inputs
        if self.execution_order is None:
            self.build()
        for op_name in self.execution_order:
            operator = self.operator_nodes[op_name]
            result = operator.execute(self.context)
            if result is not None:
                self.context.update(result)

    def __call__(self, inputs: Dict[str, Any]):
        self.reset()
        self.run(inputs)
        return self.context


class Runner:
    def __init__(self):
        self.graph = ComputationGraph()
        self.operators = {}
        self.dirty = False

    def add_operator(self, name: str, function: Callable):
        node = OperatorNode(name, function)
        self.graph.add_operator(node)
        return node

    def link_operators(self, source: OperatorNode, target: OperatorNode):
        target.input_links.append(source)

    def stack_operators(self, operators: List[OperatorNode]):
        for i in range(1, len(operators)):
            self.link_operators(operators[i-1], operators[i])

    def build_graph(self):
        if self.dirty:  
            for name, operator in self.operators.items():
                self.graph.add_operator(operator)
            self.graph.build()
            self.dirty = False
    
    def reset(self):
        self.graph.reset()

    def clear(self):
        self.graph = ComputationGraph()
        self.operators = {}
        self.dirty = False

    def run(self, inputs: Dict[str, Any]):
        self.build_graph()
        self.graph.run(inputs)

    def __call__(self, inputs: Dict[str, Any]):
        self.build_graph()
        return self.graph(inputs)



# tools/stop_music/model.py

from pydantic import BaseModel, Field
from typing import Optional, Dict, Any

class StopMusicErrorCode:
    STOP_MUSIC_FAILED = "STOP_MUSIC_FAILED"
    UNEXPECTED_ERROR = "UNEXPECTED_ERROR"
    SUCCESS = "SUCCESS"

class StopMusicParams(BaseModel):
    pass  # No parameters

class StopMusicResult(BaseModel):
    output: Optional[str] = Field(None, description="A message indicating the result.")
    error: Optional[str] = Field(None, description="Error message if any.")
    error_code: Optional[str] = Field(None, description="Error code indicating the result status.")


# toolkit/system_toolkit/clean_temp_files/api.py

import os
import shutil
from typing import Dict, Any

def clean_temp_files(paths: list = None) -> Dict[str, Any]:
    """
    Removes temporary files to free up space.

    Args:
        paths (list, optional): List of directories to clean. Defaults to system temp directories.

    Returns:
        Dict[str, Any]: A dictionary containing 'output', 'error', and 'error_code'.
    """
    try:
        if paths is None:
            import tempfile
            paths = [tempfile.gettempdir()]
        
        cleaned_paths = []
        failed_paths = {}

        for path in paths:
            if not os.path.exists(path):
                failed_paths[path] = "Path does not exist."
                continue
            if not os.path.isdir(path):
                failed_paths[path] = "Path is not a directory."
                continue
            try:
                for filename in os.listdir(path):
                    file_path = os.path.join(path, filename)
                    try:
                        if os.path.isfile(file_path) or os.path.islink(file_path):
                            os.unlink(file_path)
                        elif os.path.isdir(file_path):
                            shutil.rmtree(file_path)
                        cleaned_paths.append(file_path)
                    except Exception as e:
                        failed_paths[file_path] = str(e)
            except Exception as e:
                failed_paths[path] = str(e)
        
        return {
            "output": {
                "cleaned_files": cleaned_paths,
                "failed_files": failed_paths
            },
            "error": None if not failed_paths else "Some files/directories could not be deleted.",
            "error_code": "SUCCESS" if not failed_paths else "PARTIAL_SUCCESS"
        }
    except Exception as e:
        return {
            "output": None,
            "error": f"Failed to clean temporary files: {str(e)}",
            "error_code": "FAILED_TO_CLEAN_TEMP_FILES"
        }


# toolkit/system_toolkit/view_system_logs/test.py

import unittest
from unittest.mock import patch, MagicMock
from toolkit.system_toolkit.view_system_logs.api import view_system_logs

class TestViewSystemLogs(unittest.TestCase):
    @patch('platform.system', return_value='Linux')
    @patch('subprocess.run')
    def test_view_system_logs_success_system_linux(self, mock_run, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="System log entry 1\nSystem log entry 2", stderr="")
        response = view_system_logs("system")
        mock_run.assert_called_with(['journalctl', '-n', '100'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"]["logs"], "System log entry 1\nSystem log entry 2")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Linux')
    @patch('subprocess.run')
    def test_view_system_logs_success_application_linux(self, mock_run, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="Application log entry 1\nApplication log entry 2", stderr="")
        response = view_system_logs("application")
        mock_run.assert_called_with(['journalctl', '-u', 'application_name', '-n', '100'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"]["logs"], "Application log entry 1\nApplication log entry 2")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Linux')
    @patch('subprocess.run', side_effect=subprocess.CalledProcessError(1, ['journalctl'], stderr="Error retrieving logs."))
    def test_view_system_logs_failed_linux(self, mock_run, mock_system):
        response = view_system_logs("system")
        mock_run.assert_called_with(['journalctl', '-n', '100'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "FAILED_TO_VIEW_LOGS")
        self.assertIsNone(response["output"])
        self.assertIn("Failed to retrieve logs: Error retrieving logs.", response["error"])

    @patch('platform.system', return_value='Windows')
    @patch('subprocess.run')
    def test_view_system_logs_success_windows_system(self, mock_run, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="System log entry 1\nSystem log entry 2", stderr="")
        response = view_system_logs("system")
        mock_run.assert_called_with(['wevtutil', 'qe', 'System', '/c:100', '/f:text'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"]["logs"], "System log entry 1\nSystem log entry 2")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Windows')
    @patch('subprocess.run')
    def test_view_system_logs_success_windows_application(self, mock_run, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="Application log entry 1\nApplication log entry 2", stderr="")
        response = view_system_logs("application")
        mock_run.assert_called_with(['wevtutil', 'qe', 'Application', '/c:100', '/f:text'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"]["logs"], "Application log entry 1\nApplication log entry 2")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Windows')
    @patch('subprocess.run', side_effect=subprocess.CalledProcessError(1, ['wevtutil'], stderr="Error retrieving logs."))
    def test_view_system_logs_failed_windows(self, mock_run, mock_system):
        response = view_system_logs("system")
        mock_run.assert_called_with(['wevtutil', 'qe', 'System', '/c:100', '/f:text'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "FAILED_TO_VIEW_LOGS")
        self.assertIsNone(response["output"])
        self.assertIn("Failed to retrieve logs: Error retrieving logs.", response["error"])

    @patch('platform.system', return_value='Linux')
    def test_view_system_logs_unsupported_log_type(self, mock_system):
        response = view_system_logs("unsupported_type")
        self.assertEqual(response["error_code"], "UNSUPPORTED_LOG_TYPE")
        self.assertIsNone(response["output"])
        self.assertIn("Unsupported log type: unsupported_type", response["error"])

    @patch('platform.system', return_value='UnknownOS')
    def test_view_system_logs_unsupported_os(self, mock_system):
        response = view_system_logs("system")
        self.assertEqual(response["error_code"], "UNSUPPORTED_OS")
        self.assertIsNone(response["output"])
        self.assertIn("Unsupported operating system: UnknownOS", response["error"])

if __name__ == "__main__":
    unittest.main()


# File: cognition/action_system.py

from typing import Any, Dict, Union, List, Optional
from aeiva.action.plan import Plan
from aeiva.action.skill import Skill
from aeiva.action.task import Task
from aeiva.action.action import Action
from aeiva.tool.tool import Tool
import asyncio

import os
import json

class ActionSystem:
    """
    A concrete Action System responsible for translating Plans into executable Skills
    and managing the execution of Skills.
    """

    def __init__(self, config: Dict):
        self.config = config
        self.state = {
            "current_skill": None,
            "execution_status": "Not Started",
        }
        self.tools = []
        self.skill = None

    def setup(self) -> None:
        if "tools" in self.config.keys():
            for tool_name in self.config["tools"]:
                self.tools.append(Tool.load_tool_schema(tool_name))
        print("ActionSystem setup complete.")

    def plan_to_skill(self, plan: Plan) -> Skill:
        actions = []

        for task in plan.steps:
            if isinstance(task, Task):
                action = Action(
                    name=task.name,
                    params=task.params,
                    id=task.id,
                    dependent_ids=task.dependent_ids,
                    type="Action",
                    description=task.description,
                    metadata=task.metadata
                )
                actions.append(action)
            elif isinstance(task, Plan):
                sub_skill = self.plan_to_skill(task)  # Recursively handle sub-plans
                actions.append(sub_skill)
            else:
                raise TypeError(f"Unexpected step type: {type(task)} in plan {plan.id}")

        if not actions:
            raise ValueError(f"The plan {plan.id} does not contain any valid actions or sub-plans.")

        return Skill(
            name=plan.name,
            steps=actions,
            id=plan.id,
            dependent_ids=plan.dependent_ids,
            type="Skill",
            description=plan.description,
            metadata=plan.metadata
        )

    async def execute(self, plan: Plan) -> None:
        self.state["execution_status"] = "Executing"
        
        try:
            self.skill = self.plan_to_skill(plan)            
            await self.skill.execute()            
            self.state["execution_status"] = "Completed" if self.skill.is_successful else "Failed"
        except Exception as e:
            self.state["execution_status"] = "Failed"
            self.handle_error(e)
            raise  # Ensure to re-throw the exception

    def handle_error(self, error: Exception) -> None:
        print(f"ActionSystem encountered an error: {error}")


# tools/click_on_element/api.py

from typing import Dict, Any
import pyautogui

def click_on_element(position: Dict[str, float]) -> Dict[str, Any]:
    """
    Clicks on a GUI element at a specified position.

    Args:
        position (dict): The position dictionary containing 'x', 'y', 'width', 'height'.

    Returns:
        Dict[str, Any]: A dictionary containing 'output', 'error', and 'error_code'.
    """
    try:
        required_keys = ['x', 'y', 'width', 'height']
        for key in required_keys:
            if key not in position:
                return {
                    "output": None,
                    "error": f"Missing key '{key}' in position dictionary.",
                    "error_code": "MISSING_POSITION_KEY"
                }

        x = position['x'] + position['width'] / 2
        y = position['y'] + position['height'] / 2

        if not (0 <= x <= pyautogui.size().width) or not (0 <= y <= pyautogui.size().height):
            return {
                "output": None,
                "error": f"Position ({x}, {y}) is out of screen bounds.",
                "error_code": "POSITION_OUT_OF_BOUNDS"
            }

        pyautogui.click(x, y)
        return {
            "output": f"Clicked on element at ({x}, {y}).",
            "error": None,
            "error_code": "SUCCESS"
        }
    except Exception as e:
        return {
            "output": None,
            "error": f"Error clicking on element: {e}",
            "error_code": "CLICK_ELEMENT_FAILED"
        }


# src/aeiva/api/function/delete_file/api.py

import os
from typing import Dict, Any

async def delete_file(file_path: str, confirm: bool = False) -> Dict[str, Any]:
    """
    Deletes a specified file after confirmation.

    Args:
        file_path (str): The path to the file.
        confirm (bool): Confirmation flag to proceed with deletion.

    Returns:
        Dict[str, Any]: A dictionary containing 'output', 'error', and 'error_code'.
    """
    if not os.path.isfile(file_path):
        return {
            "output": "",
            "error": f"File not found: {file_path}",
            "error_code": "FILE_NOT_FOUND"
        }

    if not confirm:
        return {
            "output": "",
            "error": "Deletion not confirmed.",
            "error_code": "CONFIRMATION_REQUIRED"
        }

    try:
        os.remove(file_path)
        return {
            "output": f"File deleted successfully: {file_path}",
            "error": None,
            "error_code": "SUCCESS"
        }
    except Exception as e:
        return {
            "output": "",
            "error": str(e),
            "error_code": "DELETE_FAILED"
        }


# tools/scroll/test.py

import pytest
from unittest.mock import patch
from .api import scroll

@pytest.mark.asyncio
async def test_scroll_success_up():
    direction = "up"
    lines = 50
    scroll_id = 1
    expected_output = f"Scrolled {direction} by {lines} lines. Scroll ID: {scroll_id}"
    
    with patch('pyautogui.scroll') as mock_scroll:
        result = scroll(direction=direction, lines=lines, scroll_id=scroll_id)
        mock_scroll.assert_called_once_with(lines)
        assert result['output'] == expected_output
        assert result['error'] is None
        assert result['error_code'] == "SUCCESS"

@pytest.mark.asyncio
async def test_scroll_success_down():
    direction = "down"
    lines = 100
    scroll_id = 2
    expected_output = f"Scrolled {direction} by {lines} lines. Scroll ID: {scroll_id}"
    
    with patch('pyautogui.scroll') as mock_scroll:
        result = scroll(direction=direction, lines=lines, scroll_id=scroll_id)
        mock_scroll.assert_called_once_with(-lines)
        assert result['output'] == expected_output
        assert result['error'] is None
        assert result['error_code'] == "SUCCESS"

@pytest.mark.asyncio
async def test_scroll_invalid_direction():
    direction = "left"
    lines = 100
    scroll_id = 3
    expected_error = "Invalid direction. Choose 'up' or 'down'."
    expected_error_code = "INVALID_DIRECTION"
    
    result = scroll(direction=direction, lines=lines, scroll_id=scroll_id)
    assert result['output'] is None
    assert result['error'] == expected_error
    assert result['error_code'] == expected_error_code

@pytest.mark.asyncio
async def test_scroll_invalid_lines_low():
    direction = "up"
    lines = 0
    scroll_id = 4
    expected_error = "Lines must be between 1 and 1000."
    expected_error_code = "INVALID_LINES"
    
    result = scroll(direction=direction, lines=lines, scroll_id=scroll_id)
    assert result['output'] is None
    assert result['error'] == expected_error
    assert result['error_code'] == "INVALID_LINES"

@pytest.mark.asyncio
async def test_scroll_invalid_lines_high():
    direction = "down"
    lines = 1001
    scroll_id = 5
    expected_error = "Lines must be between 1 and 1000."
    expected_error_code = "INVALID_LINES"
    
    result = scroll(direction=direction, lines=lines, scroll_id=scroll_id)
    assert result['output'] is None
    assert result['error'] == expected_error
    assert result['error_code'] == "INVALID_LINES"

@pytest.mark.asyncio
async def test_scroll_exception():
    direction = "up"
    lines = 100
    scroll_id = 6
    expected_error_code = "SCROLL_FAILED"
    expected_error_message = "Error during scrolling: Sample exception."
    
    with patch('pyautogui.scroll', side_effect=Exception("Sample exception.")):
        result = scroll(direction=direction, lines=lines, scroll_id=scroll_id)
        assert result['output'] is None
        assert result['error'] == expected_error_message
        assert result['error_code'] == "SCROLL_FAILED"


import requests

class SandboxClient:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def upload_file(self, file_name: str, local_file_path: str):
        # Upload any file to the sandbox (FastAPI server)
        with open(local_file_path, 'rb') as f:
            files = {'file': (file_name, f)}
            response = requests.post(f'{self.base_url}/uploadfile/{file_name}', files=files)
        
        if response.status_code == 200:
            return response.json()
        else:
            raise ValueError(f"Error uploading file: {response.status_code}, {response.text}")

    def download_file(self, file_name: str, local_file_path: str):
        # Download any file from the sandbox (FastAPI server)
        response = requests.get(f'{self.base_url}/downloadfile/{file_name}', stream=True)
        if response.status_code == 200:
            with open(local_file_path, 'wb') as f:
                for chunk in response.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
            return {"message": f"File {file_name} downloaded successfully"}
        else:
            raise ValueError(f"Error downloading file: {response.status_code}, {response.text}")

    def execute_code(self, code: str):
        # Execute Python code inside the sandbox (FastAPI server)
        response = requests.post(f'{self.base_url}/execute_code/', json={"code": code})
        if response.status_code == 200:
            return response.json()
        else:
            raise ValueError(f"Error executing code: {response.status_code}, {response.text}")


# tools/stop_music/test.py

import pytest
from unittest.mock import patch
from .api import stop_music

@pytest.mark.asyncio
async def test_stop_music_success():
    expected_output = "Music stopped."
    
    with patch('pygame.mixer.music.stop') as mock_stop:
        result = stop_music()
        mock_stop.assert_called_once()
        assert result['output'] == expected_output
        assert result['error'] is None
        assert result['error_code'] == "SUCCESS"

@pytest.mark.asyncio
async def test_stop_music_failed():
    expected_error_code = "STOP_MUSIC_FAILED"
    expected_error_message = "Error stopping music: Sample exception."
    
    with patch('pygame.mixer.music.stop', side_effect=Exception("Sample exception.")):
        result = stop_music()
        assert result['output'] is None
        assert result['error'] == expected_error_message
        assert result['error_code'] == expected_error_code

@pytest.mark.asyncio
async def test_stop_music_unexpected_error():
    expected_error_code = "UNEXPECTED_ERROR"
    expected_error_message = "Unexpected error: Sample exception."
    
    with patch('pygame.mixer.music.stop', side_effect=Exception("Sample exception.")), \
         patch('threading.Lock.__enter__', side_effect=Exception("Sample exception.")):
        result = stop_music()
        assert result['output'] is None
        assert result['error'] == expected_error_message
        assert result['error_code'] == expected_error_code


# tools/take_screenshot/model.py

from pydantic import BaseModel, Field
from typing import Optional, Dict, Any

class TakeScreenshotErrorCode:
    MISSING_SAVE_PATH = "MISSING_SAVE_PATH"
    TAKE_SCREENSHOT_FAILED = "TAKE_SCREENSHOT_FAILED"
    SUCCESS = "SUCCESS"

class TakeScreenshotParams(BaseModel):
    save_path: Optional[str] = Field(None, description="The path to save the screenshot image. If not provided, saves to 'AI_ACCESSIBLE_PATH' with a timestamped filename.")

class TakeScreenshotResult(BaseModel):
    output: Optional[str] = Field(None, description="A message indicating the result.")
    error: Optional[str] = Field(None, description="Error message if any.")
    error_code: Optional[str] = Field(None, description="Error code indicating the result status.")


# lmp.py

import asyncio
from functools import wraps
from typing import Any, Callable, Optional, Dict, List
import os
import sys
from dotenv import load_dotenv

from aeiva.llm.llm_client import LLMClient
from aeiva.llm.llm_gateway_config import LLMGatewayConfig
from aeiva.llm.llm_gateway_exceptions import LLMGatewayError

# Load environment variables from the specified .env file
load_dotenv()
api_key = os.getenv("OPENAI_API_KEY")

if not api_key or api_key.startswith("your_ope"):
    print("Error: OPENAI_API_KEY is not set correctly in the .env file.", file=sys.stderr)
    sys.exit(1)

# Initialize LLMGatewayConfig
llm_gateway_config = LLMGatewayConfig(
    llm_model_name="gpt-4",  # Default model
    llm_api_key=api_key,
    llm_base_url="https://api.openai.com/v1",
    llm_timeout=60,
    llm_max_input_tokens=2048,
    llm_max_output_tokens=512,
    llm_temperature=0.7,
    llm_top_p=0.9,
    llm_num_retries=3,
    llm_retry_backoff_factor=2,
    llm_retry_on_status=(429, 500, 502, 503, 504),
    llm_use_async=False,
    llm_stream=False,  # Set to True for streaming responses
    llm_logging_level="INFO",
    llm_additional_params={}
)

# Initialize LLMClient
llm_client = LLMClient(llm_gateway_config)

def simple(model: Optional[str] = None, **llm_params):
    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Extract system prompt from docstring
            system_prompt = func.__doc__.strip() if func.__doc__ else ''
            # Generate user prompt
            user_prompt = func(*args, **kwargs)
            # Build messages
            messages = []
            if system_prompt:
                messages.append({'role': 'system', 'content': system_prompt})
            messages.append({'role': 'user', 'content': user_prompt})

            # Update LLM parameters
            params = {}
            if model:
                params['model'] = model
            params.update(llm_params)

            # Check if the LLM is asynchronous
            if llm_gateway_config.llm_use_async:
                # Async version
                async def async_call():
                    try:
                        # Check if streaming is enabled
                        if llm_gateway_config.llm_stream:
                            # Handle streaming response
                            response_stream = llm_client(messages=messages, **params)
                            # Collect the streamed content
                            final_response = ''
                            async for chunk in response_stream:
                                print(chunk, end='', flush=True)
                                final_response += chunk
                            print()  # Newline after streaming completes
                            return final_response
                        else:
                            response = await llm_client.agenerate(messages=messages, **params)
                            return response
                    except LLMGatewayError as e:
                        print(f"An error occurred during LLM generation: {e}", file=sys.stderr)
                        return None
                # Run the async function and return the result
                return asyncio.run(async_call())
            else:
                if llm_gateway_config.llm_stream:
                    raise NotImplementedError("Synchronous streaming is not supported.")
                else:
                    # Synchronous version
                    try:
                        response = llm_client.generate(messages=messages, **params)
                        return response
                    except LLMGatewayError as e:
                        print(f"An error occurred during LLM generation: {e}", file=sys.stderr)
                        return None
        return wrapper
    return decorator

def complex(model: Optional[str] = None, tools: Optional[List[Dict[str, Any]]] = None, **llm_params):
    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Call the original function to get messages
            result = func(*args, **kwargs)

            # Build messages
            if isinstance(result, str):
                messages = [{'role': 'user', 'content': result}]
            elif isinstance(result, list):
                messages = []
                for item in result:
                    if isinstance(item, dict):
                        messages.append(item)
                    else:
                        raise ValueError("Messages must be dictionaries with 'role' and 'content'")
            else:
                raise ValueError("Function must return a string or a list of messages")

            # Update LLM parameters
            params = {}
            if model:
                params['model'] = model
            params.update(llm_params)

            # Check if the LLM is asynchronous
            if llm_gateway_config.llm_use_async:
                # Async version
                async def async_call():
                    try:
                        # Check if streaming is enabled
                        if llm_gateway_config.llm_stream:
                            # Handle streaming response
                            response_stream = llm_client(messages=messages, tools=tools, **params)
                            # Collect the streamed content
                            final_response = ''
                            async for chunk in response_stream:
                                print(chunk, end='', flush=True)
                                final_response += chunk
                            print()  # Newline after streaming completes
                            return final_response
                        else:
                            response = await llm_client.agenerate(messages=messages, tools=tools, **params)
                            return response
                    except LLMGatewayError as e:
                        print(f"An error occurred during LLM generation: {e}", file=sys.stderr)
                        return None
                # Run the async function and return the result
                return asyncio.run(async_call())
            else:
                if llm_gateway_config.llm_stream:
                    raise NotImplementedError("Synchronous streaming is not supported.")
                else:
                    # Synchronous version
                    try:
                        response = llm_client.generate(messages=messages, tools=tools, **params)
                        return response
                    except LLMGatewayError as e:
                        print(f"An error occurred during LLM generation: {e}", file=sys.stderr)
                        return None
        return wrapper
    return decorator

# Expose llm_client for use in test.py
__all__ = ['simple', 'complex', 'llm_client']


# tools/stop_music/api.py

from typing import Dict, Any
import pygame
import threading

_lock = threading.Lock()

def stop_music() -> Dict[str, Any]:
    """
    Stops playing music.

    Returns:
        Dict[str, Any]: A dictionary containing 'output', 'error', and 'error_code'.
    """
    try:
        with _lock:
            try:
                pygame.mixer.music.stop()
                return {
                    "output": "Music stopped.",
                    "error": None,
                    "error_code": "SUCCESS"
                }
            except Exception as e:
                return {
                    "output": None,
                    "error": f"Error stopping music: {e}",
                    "error_code": "STOP_MUSIC_FAILED"
                }
    except Exception as e:
        return {
            "output": None,
            "error": f"Unexpected error: {e}",
            "error_code": "UNEXPECTED_ERROR"
        }


# test.py

import pytest
from unittest.mock import patch
from .api import rename_file
import os

@pytest.mark.asyncio
async def test_rename_file_success():
    old_file_path = "/path/to/old_file.txt"
    new_file_path = "/path/to/new_file.txt"
    expected_output = f"File renamed from {os.path.abspath(old_file_path)} to {os.path.abspath(new_file_path)}"

    with patch('os.path.exists', side_effect=lambda x: False if x == os.path.abspath(new_file_path) else True), \
         patch('os.path.abspath', side_effect=lambda x: os.path.abspath(x)), \
         patch('os.path.expanduser', side_effect=lambda x: x), \
         patch('os.makedirs') as mock_makedirs, \
         patch('os.rename') as mock_rename:
        
        result = rename_file(old_file_path, new_file_path)
        mock_makedirs.assert_not_called()
        mock_rename.assert_called_once_with(os.path.abspath(old_file_path), os.path.abspath(new_file_path))
        assert result['output'] == expected_output
        assert result['error'] is None
        assert result['error_code'] == "SUCCESS"

@pytest.mark.asyncio
async def test_rename_file_missing_paths():
    old_file_path = "   "
    new_file_path = ""
    expected_error = "Both old_file_path and new_file_path must be provided."
    expected_error_code = "MISSING_PATHS"

    result = rename_file(old_file_path, new_file_path)
    assert result['output'] is None
    assert result['error'] == expected_error
    assert result['error_code'] == "MISSING_PATHS"

@pytest.mark.asyncio
async def test_rename_file_source_not_found():
    old_file_path = "/path/to/nonexistent_file.txt"
    new_file_path = "/path/to/new_file.txt"
    expected_error = f"File not found: {os.path.abspath(old_file_path)}"
    expected_error_code = "FILE_NOT_FOUND"

    with patch('os.path.exists', return_value=False), \
         patch('os.path.abspath', side_effect=lambda x: os.path.abspath(x)), \
         patch('os.path.expanduser', side_effect=lambda x: x):
        
        result = rename_file(old_file_path, new_file_path)
        assert result['output'] is None
        assert result['error'] == expected_error
        assert result['error_code'] == "FILE_NOT_FOUND"

@pytest.mark.asyncio
async def test_rename_file_destination_exists():
    old_file_path = "/path/to/old_file.txt"
    new_file_path = "/path/to/existing_file.txt"
    expected_error = f"Destination already exists: {os.path.abspath(new_file_path)}"
    expected_error_code = "DESTINATION_EXISTS"

    with patch('os.path.exists', return_value=True), \
         patch('os.path.abspath', side_effect=lambda x: os.path.abspath(x)), \
         patch('os.path.expanduser', side_effect=lambda x: x):
        
        result = rename_file(old_file_path, new_file_path)
        assert result['output'] is None
        assert result['error'] == expected_error
        assert result['error_code'] == "DESTINATION_EXISTS"

@pytest.mark.asyncio
async def test_rename_file_permission_denied():
    old_file_path = "/protected/path/old_file.txt"
    new_file_path = "/protected/path/new_file.txt"
    expected_error = f"Permission denied: [Errno 13] Permission denied: '{os.path.abspath(new_file_path)}'"
    expected_error_code = "PERMISSION_DENIED"

    with patch('os.path.exists', return_value=False), \
         patch('os.path.abspath', side_effect=lambda x: os.path.abspath(x)), \
         patch('os.path.expanduser', side_effect=lambda x: x), \
         patch('os.makedirs', side_effect=PermissionError("Permission denied")):
        
        result = rename_file(old_file_path, new_file_path)
        assert result['output'] is None
        assert result['error'] == expected_error
        assert result['error_code'] == "PERMISSION_DENIED"

@pytest.mark.asyncio
async def test_rename_file_os_error():
    old_file_path = "/path/to/old_file.txt"
    new_file_path = "/path/to/new_file.txt"
    expected_error = f"OS error occurred: [Errno 22] Invalid argument: '{os.path.abspath(new_file_path)}'"
    expected_error_code = "OS_ERROR"

    with patch('os.path.exists', return_value=False), \
         patch('os.path.abspath', side_effect=lambda x: os.path.abspath(x)), \
         patch('os.path.expanduser', side_effect=lambda x: x), \
         patch('os.makedirs') as mock_makedirs, \
         patch('os.rename', side_effect=OSError("[Errno 22] Invalid argument")):
        
        result = rename_file(old_file_path, new_file_path)
        mock_makedirs.assert_called_once_with(os.path.dirname(os.path.abspath(new_file_path)), exist_ok=True)
        assert result['output'] is None
        assert result['error'] == expected_error
        assert result['error_code'] == "OS_ERROR"

@pytest.mark.asyncio
async def test_rename_file_unexpected_error():
    old_file_path = "/path/to/old_file.txt"
    new_file_path = "/path/to/new_file.txt"
    expected_error = f"Unexpected error: Unexpected exception"
    expected_error_code = "UNEXPECTED_ERROR"

    with patch('os.path.exists', return_value=False), \
         patch('os.path.abspath', side_effect=lambda x: os.path.abspath(x)), \
         patch('os.path.expanduser', side_effect=lambda x: x), \
         patch('os.makedirs', side_effect=Exception("Unexpected exception")):
        
        result = rename_file(old_file_path, new_file_path)
        assert result['output'] is None
        assert result['error'] == expected_error
        assert result['error_code'] == "UNEXPECTED_ERROR"


# toolkit/system_toolkit/list_drives/test.py

import unittest
from unittest.mock import patch, MagicMock
from toolkit.system_toolkit.list_drives.api import list_drives
import psutil

class TestListDrives(unittest.TestCase):
    @patch('psutil.disk_partitions')
    def test_list_drives_success(self, mock_disk_partitions):
        mock_partition1 = MagicMock()
        mock_partition1.device = '/dev/sda1'
        mock_partition1.mountpoint = '/'
        mock_partition1.fstype = 'ext4'
        mock_partition1.opts = 'rw,relatime'

        mock_partition2 = MagicMock()
        mock_partition2.device = '/dev/sda2'
        mock_partition2.mountpoint = '/home'
        mock_partition2.fstype = 'ext4'
        mock_partition2.opts = 'rw,relatime'

        mock_disk_partitions.return_value = [mock_partition1, mock_partition2]

        response = list_drives()
        mock_disk_partitions.assert_called_once_with(all=False)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"]["drives"], [
            {
                "device": '/dev/sda1',
                "mountpoint": '/',
                "fstype": 'ext4',
                "opts": 'rw,relatime'
            },
            {
                "device": '/dev/sda2',
                "mountpoint": '/home',
                "fstype": 'ext4',
                "opts": 'rw,relatime'
            }
        ])
        self.assertIsNone(response["error"])

    @patch('psutil.disk_partitions', side_effect=Exception("Mocked exception"))
    def test_list_drives_failed(self, mock_disk_partitions):
        response = list_drives()
        mock_disk_partitions.assert_called_once_with(all=False)
        self.assertEqual(response["error_code"], "FAILED_TO_LIST_DRIVES")
        self.assertIsNone(response["output"])
        self.assertIn("Failed to list drives: Mocked exception", response["error"])

if __name__ == "__main__":
    unittest.main()


# tools/click_mouse/test.py

import pytest
from unittest.mock import patch
from .api import click_mouse

@pytest.mark.asyncio
async def test_click_mouse_success():
    button = "left"
    clicks = 2
    interval = 0.1
    expected_output = "Mouse clicked 2 time(s) with left button."

    with patch('pyautogui.click') as mock_click:
        result = click_mouse(button=button, clicks=clicks, interval=interval)
        mock_click.assert_called_with(button=button, clicks=clicks, interval=interval)
        assert result['output'] == expected_output
        assert result['error'] is None
        assert result['error_code'] == "SUCCESS"

@pytest.mark.asyncio
async def test_click_mouse_invalid_button():
    button = "invalid_button"
    clicks = 1
    interval = 0.0
    expected_error = "Invalid button type: invalid_button. Choose from 'left', 'right', or 'middle'."

    result = click_mouse(button=button, clicks=clicks, interval=interval)
    assert result['output'] is None
    assert result['error'] == expected_error
    assert result['error_code'] == "INVALID_BUTTON_TYPE"

@pytest.mark.asyncio
async def test_click_mouse_invalid_clicks():
    button = "left"
    clicks = 0
    interval = 0.0
    expected_error = "Number of clicks must be at least 1."

    result = click_mouse(button=button, clicks=clicks, interval=interval)
    assert result['output'] is None
    assert result['error'] == expected_error
    assert result['error_code'] == "INVALID_CLICKS"

@pytest.mark.asyncio
async def test_click_mouse_invalid_interval():
    button = "left"
    clicks = 1
    interval = -0.5
    expected_error = "Interval between clicks cannot be negative."

    result = click_mouse(button=button, clicks=clicks, interval=interval)
    assert result['output'] is None
    assert result['error'] == expected_error
    assert result['error_code'] == "INVALID_INTERVAL"

@pytest.mark.asyncio
async def test_click_mouse_exception():
    button = "left"
    clicks = 1
    interval = 0.0
    expected_error_code = "CLICK_FAILED"
    expected_error_message = "Error clicking mouse: Sample exception."

    with patch('pyautogui.click', side_effect=Exception("Sample exception.")):
        result = click_mouse(button=button, clicks=clicks, interval=interval)
        assert result['output'] is None
        assert result['error'] == expected_error_message
        assert result['error_code'] == expected_error_code


# File: cognition/thought.py

from typing import Any, Dict, Optional
from datetime import datetime

class Thought:
    """
    Represents the output from the Brain after processing an Observation.
    """
    def __init__(self, content: Any, modality: str = 'text', timestamp: Optional[datetime] = None, metadata: Optional[Dict[str, Any]] = None):
        self.content = content  # The thought content (e.g., text)
        self.modality = modality
        self.timestamp = timestamp or datetime.now()
        self.metadata = metadata or {}

    def to_dict(self) -> Dict[str, Any]:
        return {
            'content': self.content,
            'modality': self.modality,
            'timestamp': self.timestamp.isoformat(),
            'metadata': self.metadata
        }


from aeiva.sandbox.sandbox_client import SandboxClient

# Initialize the client with the FastAPI server's URL
client = SandboxClient("http://localhost:8000")

# Test file upload
file_name = "test_upload_file.txt"
local_upload_path = "test_upload_file.txt"
try:
    upload_response = client.upload_file(file_name, local_upload_path)
    print(f"Upload Response: {upload_response}")
except Exception as e:
    print(f"Upload Error: {e}")

# Test code execution
code = """
with open('/sandbox/test_upload_file.txt', 'r') as f:
    content = f.read()
print(content)
"""
try:
    exec_response = client.execute_code(code)
    print(f"Execution Output: {exec_response['output']}")
except Exception as e:
    print(f"Execution Error: {e}")

# Test file download
local_download_path = "downloaded_test_file.txt"
try:
    download_response = client.download_file(file_name, local_download_path)
    print(f"Download Response: {download_response}")
except Exception as e:
    print(f"Download Error: {e}")


# toolkit/system_toolkit/get_user_home_path/api.py

import os
import platform
from pathlib import Path
from typing import Dict, Any

def get_user_home_path() -> Dict[str, Any]:
    """
    Retrieves the home directory of the current user across different platforms.

    Supported Platforms:
        - Windows
        - macOS
        - Linux
        - iOS (best-effort)
        - Android (best-effort)

    Returns:
        Dict[str, Any]: A dictionary containing 'output', 'error', and 'error_code'.
    """
    try:
        system = platform.system()
        print(f"Detected operating system: {system}")  # Replace with logger if needed

        if system == "Windows":
            # Windows: Use USERPROFILE or combine HOMEDRIVE and HOMEPATH
            home = os.environ.get('USERPROFILE') or os.path.join(os.environ.get('HOMEDRIVE', ''), os.environ.get('HOMEPATH', ''))
            print(f"Windows home directory: {home}")  # Replace with logger if needed
        elif system in ["Linux", "Darwin"]:  # Darwin is macOS
            # Unix-like systems: Use expanduser
            home = os.path.expanduser("~")
            print(f"Unix-like home directory: {home}")  # Replace with logger if needed
        elif system == "Java":  # Potentially Android (e.g., running via Jython or similar)
            # Android typically uses /data/user/0/<package_name>/ or /sdcard/
            # However, accessing these paths may require specific permissions
            # Here, we attempt to use the HOME environment variable
            home = os.environ.get('HOME') or '/sdcard/'
            print(f"Android home directory (best-effort): {home}")  # Replace with logger if needed
        elif system == "iOS":
            # iOS applications are sandboxed; home directory is typically the app's sandbox
            # Accessing it might require specific APIs or configurations
            # Here, we return the current working directory as a placeholder
            home = Path.cwd()
            print(f"iOS home directory (best-effort): {home}")  # Replace with logger if needed
        else:
            # Fallback for unknown systems
            home = os.path.expanduser("~")
            print(f"Unknown system '{system}'. Falling back to expanduser: {home}")  # Replace with logger if needed

        if home and os.path.isdir(home):
            return {
                "output": {"user_home": str(Path(home).resolve())},
                "error": None,
                "error_code": "SUCCESS",
            }
        else:
            return {
                "output": None,
                "error": "Determined home directory does not exist or is not a directory.",
                "error_code": "INVALID_HOME_DIRECTORY",
            }
    
    except Exception as e:
        return {
            "output": None,
            "error": f"Failed to determine the user's home directory: {str(e)}",
            "error_code": "UNEXPECTED_ERROR",
        }


from functools import partial
import os

import pytorch_lightning as pl
from pytorch_lightning.callbacks import ModelCheckpoint
import torch
from transformers import (
    AutoConfig, CLIPConfig, CLIPModel,
    LlamaModel, LlamaTokenizer, 
    WhisperConfig, WhisperForConditionalGeneration)

from aeiva.operator.dataitem_ops import (
    extract_audio_from_video, get_audio_mels, get_transformed_image,
    load_video_frames, sample_frames_from_video,
    tokenize_and_label_text_for_instruction_tuning)
from aeiva.data.loader import collate_multimodal_batches, multimodal_loader
from aeiva.model.macaw_model import MM_LLMs, MM_LLMs_Config
from aeiva.trainer.pl_trainer import LightningTrainer, LoggingCallback
from aeiva.common.constants import IGNORE_ID
from aeiva.util.file_utils import copy_file_to_dst, ensure_dir
from aeiva.common.pipeline import Pipeline
from aeiva.util.token_utils import get_tokenizer
from aeiva.common.types import TaskContext
from aeiva.common.decorators import (
    register_resource_preparer,
    register_model_loader, register_model_initializer,
    register_data_loader, register_data_processor, register_dataitem_processor,
    register_trainer, register_evaluator, register_inferer)


@register_resource_preparer("macaw")
def macaw_prepare_resource(ctx: TaskContext) -> TaskContext:
    clip_path = ctx["config"].clip_model_name_or_path
    whisper_path = ctx["config"].whisper_model_name_or_path
    llama7b_path = ctx["config"].llama7b_model_name_or_path

    # save whisper, clip, and llama models for future use.
    clip_model = CLIPModel.from_pretrained("openai/clip-vit-base-patch16")
    whisper_model = WhisperForConditionalGeneration.from_pretrained("openai/whisper-base")
    llama7b_model = LlamaModel.from_pretrained("decapoda-research/llama-7b-hf")
    ensure_dir(clip_path)
    ensure_dir(whisper_path)
    ensure_dir(llama7b_path)
    clip_model.save_pretrained(clip_path)
    whisper_model.save_pretrained(whisper_path)
    llama7b_model.save_pretrained(llama7b_path)

    return ctx


@register_model_loader("macaw")
def macaw_load_model(ctx: TaskContext) -> TaskContext:
    config = ctx["config"]

    # setup pretrained model
    clip_config = CLIPConfig.from_pretrained(config.clip_model_name_or_path)
    whisper_config = WhisperConfig.from_pretrained(config.whisper_model_name_or_path)
    llm_config = AutoConfig.from_pretrained(config.llama7b_model_name_or_path)

    tokenizer =  get_tokenizer(config.tokenizer_name_or_path, tokenizer_cls=LlamaTokenizer)
    llm_config.vocab_size = len(tokenizer)  # NOTE: because macaw model added some special tokens.

    model_config = MM_LLMs_Config(
        n_frames=6, 
        attention_heads=32, 
        image_conv_kernel=48, 
        image_conv_stride=36, 
        video_conv_kernel=36, 
        video_conv_stride=30, 
        audio_conv_kernel=240, 
        audio_conv_stride=220,
        clip_config=clip_config, whisper_config=whisper_config, llm_config=llm_config)

    # option 1: load from pretrained macaw model
    macaw_model = MM_LLMs.from_pretrained(
        config.macaw_model_name_or_path,
        config = model_config,
        # load_in_8bit=True,
        # torch_dtype=torch.float16,
        # device_map=device_map,
        )

    ctx.update({"model": macaw_model, "tokenizer": tokenizer})
    return ctx


@register_model_initializer("macaw")
def macaw_init_model(ctx: TaskContext) -> TaskContext:
    config = ctx["config"]

    # setup pretrained model
    clip_config = CLIPConfig.from_pretrained(config.clip_model_name_or_path)
    whisper_config = WhisperConfig.from_pretrained(config.whisper_model_name_or_path)
    llm_config = AutoConfig.from_pretrained(config.llama7b_model_name_or_path)

    tokenizer =  get_tokenizer(config.tokenizer_name_or_path, tokenizer_cls=LlamaTokenizer)
    llm_config.vocab_size = len(tokenizer)  # !!! because macaw model added some special tokens.

    model_config = MM_LLMs_Config(
        n_frames=6, 
        attention_heads=32, 
        image_conv_kernel=48, 
        image_conv_stride=36, 
        video_conv_kernel=36, 
        video_conv_stride=30, 
        audio_conv_kernel=240, 
        audio_conv_stride=220,
        clip_config=clip_config, whisper_config=whisper_config, llm_config=llm_config
    )

    model = MM_LLMs(config=model_config)
    model.image_encoder.from_pretrained(config.clip_model_name_or_path)
    model.video_encoder.from_pretrained(config.clip_model_name_or_path)
    model.audio_encoder.from_pretrained(config.whisper_model_name_or_path)
    model.llm.from_pretrained(config.llama7b_model_name_or_path)
    # model.llm.config.use_return_dict = True #!!!!!!!!!  # seems it is None and read only. I cannot revise it....
    model.llm.resize_token_embeddings(len(tokenizer))

    ctx.update({"model": model, "tokenizer": tokenizer})
    return ctx


@register_data_loader("macaw")
def macaw_setup_data_loader(ctx: TaskContext) -> TaskContext:
    config = ctx["config"]
    tokenizer = ctx["tokenizer"]

    data_loading_pipeline = Pipeline([
        partial(get_transformed_image, image_dir=config.image_dir),
        partial(load_video_frames, frame_dir=config.frame_dir, num_frames=config.num_frames_to_load),
        partial(get_audio_mels, audio_dir=config.audio_dir)
    ])

    if config.mode in ["inference", "infer", "eval", "evaluate", "test"]:
        config.batch_size = 1  #!!!!!!
    dataloader = multimodal_loader(config, tokenizer, data_loading_pipeline)
    print("config: ", config)

    ctx.update({"dataloader": dataloader, "config": config})  #!!! return config?

    if config.mode == "train":
        train_loader = dataloader  #!!!!!!
        val_loader = dataloader  #!!!!!!
        ctx.update({"train_loader": train_loader, "val_loader": val_loader, "data_loading_pipeline": data_loading_pipeline})
    return ctx


@register_data_processor("macaw")
def macaw_setup_data_processor(ctx: TaskContext) -> TaskContext:
    config = ctx["config"]
    tokenizer = ctx["tokenizer"]

    data_preprocess_pipeline = Pipeline([
        partial(tokenize_and_label_text_for_instruction_tuning, tokenizer=tokenizer, max_length=config.max_seq_len_for_preprocess, ignore_id=IGNORE_ID),
        partial(sample_frames_from_video, num_frames=config.num_frames_to_sample, video_dir=config.video_dir, frame_dir=config.frame_dir),  #!!! get frame indices
        partial(extract_audio_from_video, video_dir=config.video_dir, audio_dir=config.audio_dir)  # !!! get audio name
    ])
    data_loading_pipeline = Pipeline([
        partial(get_transformed_image, image_dir=config.image_dir),
        partial(load_video_frames, frame_dir=config.frame_dir, num_frames=config.num_frames_to_load),
        partial(get_audio_mels, audio_dir=config.audio_dir)
    ])

    infer_mode_data_preprocess_pipeline = Pipeline([
        partial(tokenize_and_label_text_for_instruction_tuning, tokenizer=tokenizer, max_length=config.max_seq_len_for_preprocess, ignore_id=IGNORE_ID),
        partial(sample_frames_from_video, num_frames=config.num_frames_to_sample, video_dir=config.run_time_cache_dir, frame_dir=config.run_time_cache_dir),  #!!! get frame indices
        partial(extract_audio_from_video, video_dir=config.run_time_cache_dir, audio_dir=config.run_time_cache_dir)  # !!! get audio name
    ])
    infer_mode_data_loading_pipeline = Pipeline([
        partial(get_transformed_image, image_dir=config.run_time_cache_dir),
        partial(load_video_frames, frame_dir=config.run_time_cache_dir, num_frames=config.num_frames_to_load),
        partial(get_audio_mels, audio_dir=config.run_time_cache_dir)
    ])
    
    ctx.update(
        {
            "data_preprocess_pipeline": data_preprocess_pipeline,
            "data_loading_pipeline": data_loading_pipeline,
            "infer_mode_data_preprocess_pipeline": infer_mode_data_preprocess_pipeline,
            "infer_mode_data_loading_pipeline": infer_mode_data_loading_pipeline
        }
    )
    return ctx


@register_dataitem_processor("macaw")
def macaw_process_dataitem(ctx: TaskContext) -> TaskContext:
    config = ctx["config"]
    data_preprocess_pipeline = ctx.get("infer_mode_data_preprocess_pipeline", None)
    data_loading_pipeline = ctx.get("infer_mode_data_loading_pipeline", None)

    tokenizer = ctx["tokenizer"]
    instruction = ctx["instruction"]
    input = ctx["input"]
    video_path = ctx["video_path"]
    image_path = ctx["image_path"]
    audio_path = ctx["audio_path"]

    # setup raw input item
    if image_path:
        copy_file_to_dst(image_path, config.run_time_cache_dir)
    if video_path:
        copy_file_to_dst(video_path, config.run_time_cache_dir)
    if audio_path:
        copy_file_to_dst(audio_path, config.run_time_cache_dir)

    formatted_input_e = {
        'instruction': instruction,
        'input': input,
        'output': "",
        'image': os.path.basename(image_path) if image_path else None,
        'audio': os.path.basename(video_path) if video_path else None,
        'video': os.path.basename(video_path) if video_path else None,
    }
    
    # preprocess data item
    model_input_e = data_loading_pipeline(data_preprocess_pipeline(formatted_input_e))
    batch = [model_input_e]
    data_item = collate_multimodal_batches(batch, tokenizer)

    # postprocess data item for inference
    data_item.pop("labels")  # remove these two keys for inference mode.
    data_item.pop("attention_mask")
    data_item['inference'] = True  # if not set this, the model output will be logits rather than ids.
    input_ids_list = data_item["input_ids"].tolist()[0]  # get flattened 1D list
    input_ids_list = [x for x in input_ids_list if x != tokenizer.eos_token_id and x != tokenizer.pad_token_id]
    data_item["input_ids"] = torch.tensor(input_ids_list).unsqueeze(0)

    ctx.update({"data_item": data_item})
    return ctx


@register_trainer("macaw")
def macaw_setup_trainer(ctx: TaskContext) -> TaskContext:
    config = ctx["config"]
    model = ctx["model"]
    tokenizer = ctx["tokenizer"]

    litmodel = LightningTrainer(model, tokenizer, config)

    # set up the ModelCheckpoint
    checkpoint_callback = ModelCheckpoint(
        dirpath=config.output_dir,
        filename='ckpt-{epoch:02d}-{val_loss:.2f}.pt',
        every_n_train_steps=config.save_checkpoint_every_n_train_steps,
        save_top_k=-1 if config.always_save_checkpoint else 1, #!!! set always_save_checkpoint to be False if storage is limited
        verbose=True,
        monitor='val_loss',
        mode='min',
        save_last=True  # Save the latest model, regardless of its performance. The name is last.ckpt
    )

    model_callback = LoggingCallback(
        config.log_every_n_steps,
        config.accumulate_grad_batches
    )

    # set up the Trainer
    trainer = pl.Trainer(
        # logger=wandb_logger,
        logger=None,
        callbacks=[checkpoint_callback, model_callback],
        max_steps=config.max_steps,
        accelerator='ddp' if config.ddp_backend else 'auto',
        # precision=16 if config.mixed_precision else 32,
        precision=32,
        deterministic=True,
        gradient_clip_val=config.max_grad_norm,
        accumulate_grad_batches=config.accumulate_grad_batches,
        val_check_interval=config.val_check_interval, #!!! this may make the training slow as we eval frequently. disable it to eval only after each epoch
        limit_val_batches=config.limit_val_batches, #!!! Not using full validation set. Disable it if evaluate after each epoch
        log_every_n_steps=config.log_every_n_steps
    )

    ctx.update({"trainer": trainer, "model": litmodel})
    return ctx


@register_evaluator("macaw")
def macaw_eval(ctx: TaskContext) -> TaskContext:
    dataloader = ctx["dataloader"]
    model = ctx["model"]
    tokenizer = ctx["tokenizer"]
    max_samples_for_inference = 3 #!!!!!! changeg to config param later.

    num_samples = 0
    model.eval()
    with torch.no_grad():
        for data_item in dataloader:
            if num_samples >= max_samples_for_inference:
                break
            # postprocess data item for inference
            data_item.pop("labels")  #!!! remove these two keys for inference mode.
            data_item.pop("attention_mask")
            data_item['inference'] = True  #!!! if not set this, the model output will be logits rather than ids.
            eos_token_id = tokenizer.eos_token_id
            pad_token_id = tokenizer.pad_token_id
            input_ids_list = data_item["input_ids"].tolist()[0]  #!!! get flattened 1D list
            input_ids_list = [x for x in input_ids_list if x != eos_token_id and x != pad_token_id]
            data_item["input_ids"] = torch.tensor(input_ids_list).unsqueeze(0)
            #!!! we need to also remove the Response part of the item. revise later.

            generate_ids = model(data_item)
            generated_texts = tokenizer.batch_decode(generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False)
            input_texts = tokenizer.batch_decode(data_item["input_ids"], skip_special_tokens=True, clean_up_tokenization_spaces=False)
            print("input_texts: ", input_texts)
            print("generated_texts: ", generated_texts)
            num_samples += 1
        #!!!! NOTE: ctx is not updated yet. We may record the inference results in ctx here. revise later.
    return ctx


@register_inferer("macaw")
def macaw_infer(ctx: TaskContext) -> TaskContext:
    data_item = ctx["data_item"]
    model = ctx["model"]
    tokenizer = ctx["tokenizer"]

    model.eval()
    with torch.no_grad():
        generate_ids = model(data_item)
    generated_texts = tokenizer.batch_decode(generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False)[0]

    ctx.update({"generated_texts": generated_texts})
    return ctx



# toolkit/file_toolkit/extract_file_as_markdown/test.py

import unittest
from unittest.mock import patch, MagicMock
from toolkit.file_toolkit.extract_file_as_markdown.api import (
    pdf2markdown
)
import os

# Import WebDriverException if used in api.py
from selenium.common.exceptions import WebDriverException


class TestPdf2Markdown(unittest.TestCase):
    @patch('pymupdf4llm.to_markdown')
    @patch('builtins.open')
    @patch('os.path.isfile', return_value=True)
    def test_pdf2markdown_success_return_markdown(self, mock_isfile, mock_open, mock_to_markdown):
        # Mock the to_markdown function to return sample markdown
        mock_to_markdown.return_value = "# Sample Markdown"

        response = pdf2markdown(
            input_file_path="sample.pdf",
            pages=None,
            output_file_path=None
        )

        # Assertions
        mock_isfile.assert_called_once_with("sample.pdf")
        mock_to_markdown.assert_called_once_with("sample.pdf", None)
        mock_open.assert_not_called()

        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertIsNone(response["error"])
        self.assertEqual(response["output"]["markdown"], "# Sample Markdown")

    @patch('pymupdf4llm.to_markdown')
    @patch('builtins.open', new_callable=unittest.mock.mock_open)
    @patch('os.path.isfile', return_value=True)
    def test_pdf2markdown_success_save_to_file(self, mock_isfile, mock_open, mock_to_markdown):
        # Mock the to_markdown function to return sample markdown
        mock_to_markdown.return_value = "# Sample Markdown"

        response = pdf2markdown(
            input_file_path="sample.pdf",
            pages=[0, 1],
            output_file_path="output.md"
        )

        # Assertions
        mock_isfile.assert_called_once_with("sample.pdf")
        mock_to_markdown.assert_called_once_with("sample.pdf", [0, 1])
        mock_open.assert_called_once_with("output.md", 'w', encoding='utf-8')
        mock_open().write.assert_called_once_with("# Sample Markdown")

        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertIsNone(response["error"])
        self.assertEqual(response["output"]["message"], "Markdown file saved to 'output.md'.")

    @patch('os.path.isfile', return_value=False)
    def test_pdf2markdown_file_not_found(self, mock_isfile):
        response = pdf2markdown(
            input_file_path="nonexistent.pdf",
            pages=None,
            output_file_path=None
        )

        # Assertions
        mock_isfile.assert_called_once_with("nonexistent.pdf")

        self.assertEqual(response["error_code"], "FILE_NOT_FOUND")
        self.assertIsNone(response["output"])
        self.assertIn("Input file 'nonexistent.pdf' does not exist.", response["error"])

    @patch('pymupdf4llm.to_markdown', side_effect=Exception("JavaScript Error"))
    @patch('os.path.isfile', return_value=True)
    def test_pdf2markdown_js_execution_error(self, mock_isfile, mock_to_markdown):
        response = pdf2markdown(
            input_file_path="sample.pdf",
            pages=None,
            output_file_path=None
        )

        # Assertions
        mock_isfile.assert_called_once_with("sample.pdf")
        mock_to_markdown.assert_called_once_with("sample.pdf", None)

        self.assertEqual(response["error_code"], "JS_EXECUTION_ERROR")
        self.assertIsNone(response["output"])
        self.assertIn("JS_EXECUTION_ERROR: JavaScript Error", response["error"])

    @patch('builtins.open', side_effect=IOError("Failed to write file"))
    @patch('pymupdf4llm.to_markdown')
    @patch('os.path.isfile', return_value=True)
    def test_pdf2markdown_io_error(self, mock_isfile, mock_to_markdown, mock_open):
        # Mock the to_markdown function to return sample markdown
        mock_to_markdown.return_value = "# Sample Markdown"

        response = pdf2markdown(
            input_file_path="sample.pdf",
            pages=[0],
            output_file_path="output.md"
        )

        # Assertions
        mock_isfile.assert_called_once_with("sample.pdf")
        mock_to_markdown.assert_called_once_with("sample.pdf", [0])
        mock_open.assert_called_once_with("output.md", 'w', encoding='utf-8')

        self.assertEqual(response["error_code"], "IO_ERROR")
        self.assertIsNone(response["output"])
        self.assertIn("IO_ERROR: Failed to write file", response["error"])

    @patch('pymupdf4llm.to_markdown', side_effect=WebDriverException("WebDriver failed"))
    @patch('os.path.isfile', return_value=True)
    def test_pdf2markdown_webdriver_error(self, mock_isfile, mock_to_markdown):
        response = pdf2markdown(
            input_file_path="sample.pdf",
            pages=None,
            output_file_path=None
        )

        # Assertions
        mock_isfile.assert_called_once_with("sample.pdf")
        mock_to_markdown.assert_called_once_with("sample.pdf", None)

        self.assertEqual(response["error_code"], "JS_EXECUTION_ERROR")
        self.assertIsNone(response["output"])
        self.assertIn("JS_EXECUTION_ERROR: WebDriver failed", response["error"])

    @patch('pymupdf4llm.to_markdown', side_effect=Exception("Unexpected Error"))
    @patch('os.path.isfile', return_value=True)
    def test_pdf2markdown_unexpected_error(self, mock_isfile, mock_to_markdown):
        response = pdf2markdown(
            input_file_path="sample.pdf",
            pages=None,
            output_file_path=None
        )

        # Assertions
        mock_isfile.assert_called_once_with("sample.pdf")
        mock_to_markdown.assert_called_once_with("sample.pdf", None)

        self.assertEqual(response["error_code"], "UNEXPECTED_ERROR")
        self.assertIsNone(response["output"])
        self.assertIn("UNEXPECTED_ERROR: Unexpected Error", response["error"])

    def test_pdf2markdown_validation_error_missing_params(self):
        # Missing required 'input_file_path' parameter
        with self.assertRaises(TypeError):
            # This should raise a TypeError before even entering the function
            pdf2markdown(
                input_file_path=None,  # Invalid input
                pages=None,
                output_file_path=None
            )

    @patch('pymupdf4llm.to_markdown')
    @patch('builtins.open', new_callable=unittest.mock.mock_open)
    @patch('os.path.isfile', return_value=True)
    def test_pdf2markdown_empty_pages_list(self, mock_isfile, mock_open, mock_to_markdown):
        # Mock the to_markdown function to return sample markdown
        mock_to_markdown.return_value = "# All Pages Markdown"

        response = pdf2markdown(
            input_file_path="sample.pdf",
            pages=[],
            output_file_path=None
        )

        # Assertions
        mock_isfile.assert_called_once_with("sample.pdf")
        mock_to_markdown.assert_called_once_with("sample.pdf", [])
        mock_open.assert_not_called()

        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertIsNone(response["error"])
        self.assertEqual(response["output"]["markdown"], "# All Pages Markdown")


if __name__ == "__main__":
    unittest.main()


from typing import List, Dict, Any, Optional, Union
from aeiva.action.procedure import Procedure
from aeiva.action.task import Task

class Plan(Procedure):
    """
    Represents a plan, which is a structured roadmap for achieving a goal by executing tasks and sub-plans.
    Inherits common functionality from Procedure.
    """

    def __init__(self, name: str, steps: List[Union['Plan', Task]],
                 id: Optional[str] = None, dependent_ids: Optional[List[str]] = None,
                 type: Optional[str] = None, description: Optional[str] = None,
                 metadata: Optional[Dict[str, Any]] = None):
        """
        Initializes a Skill by extending Procedure.
        """
        super().__init__(name=name, steps=steps,
                         id=id, dependent_ids=dependent_ids,
                         type=type, description=description,
                         metadata=metadata)
        self.type = "Plan"


# tools/close_application/model.py

from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any

class CloseApplicationErrorCode:
    MISSING_PROCESS_NAME = "MISSING_PROCESS_NAME"
    PROCESS_NOT_FOUND = "PROCESS_NOT_FOUND"
    CLOSE_APPLICATION_FAILED = "CLOSE_APPLICATION_FAILED"
    SUCCESS = "SUCCESS"

class CloseApplicationParams(BaseModel):
    process_name: str = Field(..., description="The name of the process to terminate.")

class CloseApplicationResult(BaseModel):
    output: Optional[str] = Field(None, description="A message indicating the result.")
    error: Optional[str] = Field(None, description="Error message if any.")
    error_code: Optional[str] = Field(None, description="Error code indicating the result status.")


# refresh_webpage/test.py

import pytest
from unittest.mock import patch, MagicMock
from refresh_webpage.api import refresh_webpage
from selenium.common.exceptions import WebDriverException

@pytest.mark.parametrize("request, mock_refresh_side_effect, expected", [
    # Successful refresh without ignoring cache
    (
        {
            "ignore_cache": False
        },
        None,  # No exception
        {
            "output": {
                "success": True,
                "previous_url": "http://example.com",
                "new_url": None
            },
            "error": None,
            "error_code": "SUCCESS"
        }
    ),
    # Successful refresh with ignoring cache
    (
        {
            "ignore_cache": True
        },
        None,
        {
            "output": {
                "success": True,
                "previous_url": "http://example.com",
                "new_url": None
            },
            "error": None,
            "error_code": "SUCCESS"
        }
    ),
    # Missing optional parameter (use defaults)
    (
        {},
        None,
        {
            "output": {
                "success": True,
                "previous_url": "http://example.com",
                "new_url": None
            },
            "error": None,
            "error_code": "SUCCESS"
        }
    ),
    # WebDriver exception during refresh
    (
        {
            "ignore_cache": False
        },
        WebDriverException("Refresh failed"),
        {
            "output": None,
            "error": "WebDriver Error during refresh: Refresh failed",
            "error_code": "WEBDRIVER_ERROR"
        }
    ),
    # Unexpected exception
    (
        {
            "ignore_cache": False
        },
        "Some unexpected error",
        {
            "output": None,
            "error": "Unexpected Error: Some unexpected error",
            "error_code": "UNEXPECTED_ERROR"
        }
    ),
])
def test_refresh_webpage(request, mock_refresh_side_effect, expected):
    with patch('refresh_webpage.api.webdriver.Chrome') as mock_webdriver:
        mock_driver = MagicMock()
        mock_webdriver.return_value = mock_driver

        if mock_refresh_side_effect == "Some unexpected error":
            mock_driver.refresh.side_effect = Exception(mock_refresh_side_effect)
        elif isinstance(mock_refresh_side_effect, WebDriverException):
            mock_driver.refresh.side_effect = mock_refresh_side_effect
        else:
            mock_driver.refresh.return_value = None
            mock_driver.execute_script.return_value = None

        # Mock driver.current_url
        mock_driver.current_url = "http://example.com"

        # Mock driver.quit
        mock_driver.quit.return_value = None

        # Execute
        result = refresh_webpage(request)

        # Assert
        assert result == expected


# File: cognition/emotion_componential.py

from dataclasses import dataclass, field
from typing import Dict, Any
import json
import logging

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

@dataclass
class ComponentialEmotionState:
    """
    Represents the emotional state based on the Componential Model.
    
    Attributes:
        emotion_label (str): Current emotion category.
        intensity (float): Intensity of the emotion (0.0 to 1.0).
    """
    emotion_label: str = "neutral"
    intensity: float = 0.0

    def to_dict(self) -> Dict[str, Any]:
        return {
            'emotion_label': self.emotion_label,
            'intensity': self.intensity
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        return ComponentialEmotionState(
            emotion_label=data.get('emotion_label', 'neutral'),
            intensity=data.get('intensity', 0.0)
        )

from typing import Dict, Any
from aeiva.cognition.emotion.emotion import Emotion, ConfigurationError, UpdateError, RegulationError
import asyncio

class ComponentialEmotion(Emotion[ComponentialEmotionState]):
    """
    Concrete implementation of the Emotion base class using the Componential Model.
    """
    
    # Define possible emotions
    EMOTIONS = {
        'joy', 'sadness', 'anger', 'fear', 'surprise', 'disgust',
        'trust', 'anticipation', 'contentment', 'frustration'
    }
    
    def init_state(self) -> ComponentialEmotionState:
        default_emotion = self.config.get('default_emotion', 'neutral')
        default_intensity = self.config.get('default_intensity', 0.0)
        if default_emotion not in self.EMOTIONS and default_emotion != 'neutral':
            raise ConfigurationError(f"Invalid default emotion: {default_emotion}")
        return ComponentialEmotionState(emotion_label=default_emotion, intensity=default_intensity)
    
    async def setup(self) -> None:
        """
        Set up the Componential Emotion system.
        """
        required_keys = ['appraisal_weights', 'decay_rate']
        for key in required_keys:
            if key not in self.config:
                raise ConfigurationError(f"Missing configuration parameter: {key}")
        
        self.appraisal_weights = self.config['appraisal_weights']  # e.g., {'relevance': 0.3, ...}
        self.decay_rate = self.config['decay_rate']
        
        logging.info("ComponentialEmotion setup complete.")
    
    async def update(self, input_data: Dict[str, Any]) -> None:
        """
        Update the emotional state based on input stimuli via appraisal.
        """
        try:
            # Extract appraisal factors
            relevance = input_data.get('relevance', 0.0) * self.appraisal_weights.get('relevance', 0.0)
            implication = input_data.get('implication', 0.0) * self.appraisal_weights.get('implication', 0.0)
            causality = input_data.get('causality', 0.0) * self.appraisal_weights.get('causality', 0.0)
            control = input_data.get('control', 0.0) * self.appraisal_weights.get('control', 0.0)
            future_expectation = input_data.get('future_expectation', 0.0) * self.appraisal_weights.get('future_expectation', 0.0)
            
            # Simple mapping logic to determine emotion based on appraisal
            total = relevance + implication + causality + control + future_expectation
            
            if total > 2.0:
                emotion = 'joy'
            elif 1.5 < total <= 2.0:
                emotion = 'anticipation'
            elif 1.0 < total <= 1.5:
                emotion = 'trust'
            elif 0.5 < total <= 1.0:
                emotion = 'surprise'
            else:
                emotion = 'sadness'
            
            self.state.emotion_label = emotion
            self.state.intensity = min(total / 3.0, 1.0)  # Normalize intensity
            
            logging.debug(f"Appraisal results: {total}, Mapped Emotion: {emotion}, Intensity: {self.state.intensity}")
        except Exception as e:
            self.handle_error(e)
            raise UpdateError("Failed to update emotional state via appraisal.") from e
    
    def regulate(self, strategy: str) -> None:
        """
        Regulate the emotional state using a specified strategy.
        """
        try:
            if strategy == 'decay':
                self.state.intensity = max(self.state.intensity - self.decay_rate, 0.0)
                if self.state.intensity == 0.0:
                    self.state.emotion_label = 'neutral'
                logging.info("Applied regulation strategy: decay")
            elif strategy == 'reset':
                self.state.emotion_label = 'neutral'
                self.state.intensity = 0.0
                logging.info("Applied regulation strategy: reset")
            else:
                raise ValueError(f"Unknown regulation strategy: {strategy}")
        except Exception as e:
            self.handle_error(e)
            raise RegulationError(f"Failed to apply regulation strategy: {strategy}") from e
    
    def express(self) -> str:
        """
        Generate a textual expression of the current emotional state.
        """
        if self.state.emotion_label == 'neutral' or self.state.intensity == 0.0:
            return "I'm feeling neutral."
        else:
            intensity_desc = ""
            if self.state.intensity > 0.7:
                intensity_desc = " intensely"
            elif self.state.intensity > 0.4:
                intensity_desc = ""
            else:
                intensity_desc = " slightly"
            return f"I feel{intensity_desc} {self.state.emotion_label}."
    
    def serialize(self) -> str:
        return json.dumps(self.state.to_dict())
    
    def deserialize(self, data: str) -> None:
        try:
            state_dict = json.loads(data)
            self.state = ComponentialEmotionState.from_dict(state_dict)
            logging.info("Deserialized Componential emotional state.")
        except Exception as e:
            self.handle_error(e)
            raise UpdateError("Failed to deserialize Componential emotional state.") from e

# File: main_componential.py

import asyncio
from aeiva.cognition.emotion.emotion_componential import ComponentialEmotion

async def main():
    # Configuration for ComponentialEmotion
    config = {
        'appraisal_weights': {
            'relevance': 0.5,
            'implication': 0.4,
            'causality': 0.3,
            'control': 0.2,
            'future_expectation': 0.1
        },
        'decay_rate': 0.2,  # Intensity decreases by 0.2 on decay
        'default_emotions': {
            'emotion_label': 'neutral',
            'intensity': 0.0
        }
    }

    # Instantiate and set up the emotion system
    emotion_system = ComponentialEmotion(config)
    await emotion_system.setup()

    # Express initial state
    print(emotion_system.express())  # Output: I'm feeling neutral.

    # Update with stimuli: high relevance and implication
    await emotion_system.update({
        'relevance': 2.0,
        'implication': 1.5,
        'causality': 0.5,
        'control': 0.3,
        'future_expectation': 0.2
    })
    print(emotion_system.express())  # Output: I feel joy. or similar based on mapping

    # Apply regulation: decay
    emotion_system.regulate('decay')
    print(emotion_system.express())  # Output may vary based on updated intensity

    # Apply regulation: reset
    emotion_system.regulate('reset')
    print(emotion_system.express())  # Output: I'm feeling neutral.

    # Serialize the emotional state
    serialized_state = emotion_system.serialize()
    print(f"Serialized State: {serialized_state}")

    # Deserialize into a new emotion system instance
    new_emotion_system = ComponentialEmotion(config)
    await new_emotion_system.setup()
    new_emotion_system.deserialize(serialized_state)
    print(new_emotion_system.express())  # Output matches the serialized state

if __name__ == '__main__':
    asyncio.run(main())


# aeiva/plugin/plug.py

"""
Plug Module
-----------
This module provides a flexible plugin system with support for:

- Multiple plugin sources with isolation
- Context managers and import hooks
- Resource loading from plugins
- Loading plugins from directories and zip files
- Hot swapping and lazy loading of plugins

Author: Bang Liu
Date: 2024-11-19
"""

import abc
import os
import sys
import threading
import importlib
import importlib.util
import importlib.abc
import zipfile
from types import ModuleType
from typing import List, Optional, Dict


class Plugin(abc.ABC):
    """
    Abstract base class that all plugins must inherit from.
    """

    @abc.abstractmethod
    def activate(self) -> None:
        """Method called when the plugin is activated."""
        pass

    @abc.abstractmethod
    def deactivate(self) -> None:
        """Method called when the plugin is deactivated."""
        pass


class PluginLoader(importlib.abc.Loader):
    """
    Custom loader for plugin modules.
    Loads the `plugin.py` file within the plugin directory.
    """

    def __init__(self, plugin_source: 'PluginSource', plugin_name: str) -> None:
        self.plugin_source = plugin_source
        self.plugin_name = plugin_name

    def create_module(self, spec: importlib.machinery.ModuleSpec) -> Optional[ModuleType]:
        """Use default module creation semantics."""
        return None

    def exec_module(self, module: ModuleType) -> None:
        """Execute the plugin's `plugin.py` module."""
        try:
            code = self.plugin_source.get_plugin_code(self.plugin_name)
        except ImportError as e:
            print(f"PluginLoader: Failed to get code for plugin '{self.plugin_name}': {e}")
            raise

        # Compute project_root dynamically based on plug.py's location
        plugin_dir = os.path.dirname(os.path.abspath(__file__))
        project_root = os.path.abspath(os.path.join(plugin_dir, '../../../'))
        print(f"PluginLoader: Adding '{project_root}' to sys.path for plugin '{self.plugin_name}'")
        sys.path.insert(0, project_root)

        try:
            print(f"PluginLoader: Executing plugin '{self.plugin_name}'")
            exec(code, module.__dict__)
            print(f"PluginLoader: Plugin '{self.plugin_name}' executed successfully")
        except Exception as e:
            print(f"PluginLoader: Error executing plugin '{self.plugin_name}': {e}")
            raise
        finally:
            sys.path.pop(0)


class PluginFinder(importlib.abc.MetaPathFinder):
    """
    Custom finder for plugin modules.
    Finds plugins as directories containing a `plugin.py` file.
    """

    def __init__(self, plugin_source: 'PluginSource') -> None:
        self.plugin_source = plugin_source

    def find_spec(
        self,
        fullname: str,
        path: Optional[List[str]],
        target: Optional[ModuleType] = None
    ) -> Optional[importlib.machinery.ModuleSpec]:
        """
        Find the module spec for the given module.
        Handles both the namespace package and its submodules (plugins).
        """
        if fullname == self.plugin_source.namespace:
            # Handle the namespace package itself
            print(f"PluginFinder: Creating namespace package '{fullname}'")
            spec = importlib.machinery.ModuleSpec(fullname, loader=None, is_package=True)
            spec.submodule_search_locations = []
            return spec

        elif fullname.startswith(self.plugin_source.namespace + '.'):
            # Handle submodules (plugins)
            plugin_name = fullname[len(self.plugin_source.namespace) + 1:]
            if plugin_name in self.plugin_source.list_plugins():
                print(f"PluginFinder: Found plugin '{plugin_name}' for module '{fullname}'")
                loader = PluginLoader(self.plugin_source, plugin_name)
                spec = importlib.util.spec_from_loader(fullname, loader)
                spec.submodule_search_locations = []
                return spec

        # If not handling this module, return None
        print(f"PluginFinder: Not handling module '{fullname}'")
        return None


class PluginSource:
    """
    Represents an isolated source of plugins.
    Each plugin is a directory containing a `plugin.py` file.
    """

    def __init__(self, name: str, search_path: Optional[List[str]] = None) -> None:
        """
        Initializes the PluginSource.

        :param name: Unique name for the plugin source.
        :param search_path: List of paths (directories or zip files) to search for plugins.
        """
        self.name = name
        self.search_path = search_path or []
        self._lock = threading.Lock()
        self._modules: Dict[str, ModuleType] = {}
        self.namespace = f"_plug_{self.name}"
        self._finder = PluginFinder(self)
        self._finder_enabled = False

    def __enter__(self) -> 'PluginSource':
        """Enter the runtime context related to this object."""
        self.enable()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the runtime context."""
        self.disable()

    def enable(self) -> None:
        """Enable the plugin import mechanism."""
        if not self._finder_enabled:
            sys.meta_path.insert(0, self._finder)
            self._finder_enabled = True
            print(f"PluginSource: Import hook enabled for namespace '{self.namespace}'.")

    def disable(self) -> None:
        """Disable the plugin import mechanism."""
        if self._finder_enabled:
            try:
                sys.meta_path.remove(self._finder)
                print(f"PluginSource: Import hook disabled for namespace '{self.namespace}'.")
            except ValueError:
                print(f"PluginSource: Import hook for namespace '{self.namespace}' was not found in sys.meta_path.")
            self._finder_enabled = False

    def list_plugins(self) -> List[str]:
        """
        Lists available plugins in the search paths.
        Each plugin is a directory containing a `plugin.py` file.

        :return: List of plugin names.
        """
        plugins = set()
        for path in self.search_path:
            if zipfile.is_zipfile(path):
                with zipfile.ZipFile(path, 'r') as z:
                    # Identify top-level directories containing `plugin.py`
                    plugin_dirs = set()
                    for file in z.namelist():
                        parts = file.split('/')
                        if len(parts) >= 2 and parts[-1] == 'plugin.py':
                            plugin_dir = parts[0]
                            plugin_dirs.add(plugin_dir)
                    plugins.update(plugin_dirs)
            else:
                # Assume it's a directory
                if not os.path.isdir(path):
                    print(f"PluginSource: Path '{path}' is not a directory or a zip file. Skipping.")
                    continue
                for entry in os.listdir(path):
                    plugin_path = os.path.join(path, entry)
                    if os.path.isdir(plugin_path):
                        plugin_main = os.path.join(plugin_path, 'plugin.py')
                        if os.path.isfile(plugin_main):
                            plugins.add(entry)
        return list(plugins)

    def get_plugin_code(self, plugin_name: str) -> str:
        """
        Get the source code of the plugin's `plugin.py`.

        :param plugin_name: Name of the plugin to load.
        :return: Source code of `plugin.py` as a string.
        """
        for path in self.search_path:
            if zipfile.is_zipfile(path):
                with zipfile.ZipFile(path, 'r') as z:
                    plugin_main = f"{plugin_name}/plugin.py"
                    if plugin_main in z.namelist():
                        print(f"PluginSource: Found plugin '{plugin_name}' in zip file '{path}'.")
                        return z.read(plugin_main).decode('utf-8')
            else:
                # Assume it's a directory
                plugin_dir = os.path.join(path, plugin_name)
                plugin_main = os.path.join(plugin_dir, 'plugin.py')
                if os.path.isfile(plugin_main):
                    print(f"PluginSource: Found plugin '{plugin_name}' as module file '{plugin_main}'.")
                    with open(plugin_main, 'r', encoding='utf-8') as f:
                        return f.read()
        raise ImportError(f"Cannot find plugin '{plugin_name}'.")

    def load_plugin(self, plugin_name: str) -> ModuleType:
        """
        Loads a plugin by name.

        :param plugin_name: Name of the plugin to load.
        :return: The loaded plugin module.
        """
        with self._lock:
            full_name = f"{self.namespace}.{plugin_name}"
            if full_name in sys.modules:
                print(f"PluginSource: Plugin '{plugin_name}' is already loaded as '{full_name}'.")
                return sys.modules[full_name]
            # Enable the finder if not already enabled
            self.enable()
            try:
                print(f"PluginSource: Loading plugin '{plugin_name}' as '{full_name}'.")
                module = importlib.import_module(full_name)
                self._modules[plugin_name] = module
                return module
            except ImportError as e:
                print(f"PluginSource: Cannot import plugin '{plugin_name}': {e}")
                raise

    def unload_plugin(self, plugin_name: str) -> None:
        """
        Unloads a plugin by name.

        :param plugin_name: Name of the plugin to unload.
        """
        with self._lock:
            full_name = f"{self.namespace}.{plugin_name}"
            module = self._modules.pop(plugin_name, None)
            if module:
                if hasattr(module, 'deactivate'):
                    try:
                        print(f"PluginSource: Deactivating plugin '{plugin_name}'.")
                        getattr(module, 'deactivate')()
                    except Exception as e:
                        print(f"PluginSource: Error during deactivation of plugin '{plugin_name}': {e}")
                if full_name in sys.modules:
                    del sys.modules[full_name]
                    print(f"PluginSource: Plugin '{plugin_name}' unloaded and removed from sys.modules.")
            else:
                print(f"PluginSource: Plugin '{plugin_name}' is not loaded.")

    def load_resource(self, plugin_name: str, resource_name: str) -> bytes:
        """
        Loads a resource from a plugin.

        :param plugin_name: Name of the plugin.
        :param resource_name: Name of the resource file.
        :return: Contents of the resource file as bytes.
        """
        for path in self.search_path:
            if zipfile.is_zipfile(path):
                with zipfile.ZipFile(path, 'r') as z:
                    resource_file = f"{plugin_name}/{resource_name}"
                    if resource_file in z.namelist():
                        print(f"PluginSource: Loading resource '{resource_name}' from plugin '{plugin_name}' in zip '{path}'.")
                        return z.read(resource_file)
            else:
                # Assume it's a directory
                resource_path = os.path.join(path, plugin_name, resource_name)
                if os.path.isfile(resource_path):
                    print(f"PluginSource: Loading resource '{resource_name}' from plugin '{plugin_name}' at '{resource_path}'.")
                    with open(resource_path, 'rb') as f:
                        return f.read()
        raise FileNotFoundError(f"Resource '{resource_name}' not found in plugin '{plugin_name}'.")


class PluginManager:
    """
    Manages multiple PluginSources and controls plugin imports.
    """

    def __init__(self) -> None:
        self.plugin_sources: Dict[str, PluginSource] = {}

    def create_plugin_source(self, name: str, search_path: Optional[List[str]] = None) -> PluginSource:
        """
        Creates a new PluginSource.

        :param name: Unique name for the plugin source.
        :param search_path: List of paths to search for plugins.
        :return: The created PluginSource.
        """
        if name in self.plugin_sources:
            raise ValueError(f"Plugin source '{name}' already exists.")
        source = PluginSource(name, search_path)
        self.plugin_sources[name] = source
        print(f"PluginManager: Created plugin source '{name}' with search paths {search_path}.")
        return source

    def get_plugin_source(self, name: str) -> Optional[PluginSource]:
        """
        Retrieves a PluginSource by name.

        :param name: Name of the PluginSource.
        :return: The PluginSource instance, or None if not found.
        """
        return self.plugin_sources.get(name)

    def remove_plugin_source(self, name: str) -> None:
        """
        Removes a PluginSource.

        :param name: Name of the PluginSource to remove.
        """
        source = self.plugin_sources.pop(name, None)
        if source:
            source.disable()
            for plugin_name in list(source._modules.keys()):
                source.unload_plugin(plugin_name)
            print(f"PluginManager: Removed plugin source '{name}'.")
        else:
            print(f"PluginManager: Plugin source '{name}' does not exist.")


# toolkit/system_toolkit/update_system_packages/test.py

import unittest
from unittest.mock import patch, MagicMock
from toolkit.system_toolkit.update_system_packages.api import update_system_packages

class TestUpdateSystemPackages(unittest.TestCase):
    @patch('platform.system', return_value='Windows')
    @patch('subprocess.run')
    def test_update_system_packages_success_windows(self, mock_run, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="All packages upgraded successfully.", stderr="")
        response = update_system_packages()
        mock_run.assert_called_with('winget upgrade --all', shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"], "All packages upgraded successfully.")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Darwin')
    @patch('subprocess.run')
    def test_update_system_packages_success_mac(self, mock_run, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="brew update && brew upgrade executed successfully.", stderr="")
        response = update_system_packages()
        mock_run.assert_called_with('brew update && brew upgrade', shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"], "brew update && brew upgrade executed successfully.")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Linux')
    @patch('platform.linux_distribution', return_value=('Ubuntu', '20.04', 'focal'))
    @patch('subprocess.run')
    def test_update_system_packages_success_linux_ubuntu(self, mock_run, mock_distro, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="sudo apt-get update && sudo apt-get upgrade -y executed successfully.", stderr="")
        response = update_system_packages()
        mock_run.assert_called_with('sudo apt-get update && sudo apt-get upgrade -y', shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"], "sudo apt-get update && sudo apt-get upgrade -y executed successfully.")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Linux')
    @patch('platform.linux_distribution', return_value=('Fedora', '33', ''))
    @patch('subprocess.run')
    def test_update_system_packages_success_linux_fedora(self, mock_run, mock_distro, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="sudo dnf upgrade -y executed successfully.", stderr="")
        response = update_system_packages()
        mock_run.assert_called_with('sudo dnf upgrade -y', shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"], "sudo dnf upgrade -y executed successfully.")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Linux')
    @patch('platform.linux_distribution', return_value=('Arch', 'rolling', ''))
    def test_update_system_packages_unsupported_distribution(self, mock_distro, mock_system):
        response = update_system_packages()
        self.assertEqual(response["error_code"], "UNSUPPORTED_DISTRIBUTION")
        self.assertIsNone(response["output"])
        self.assertIn("Unsupported Linux distribution: arch", response["error"])

    @patch('platform.system', return_value='UnknownOS')
    def test_update_system_packages_unsupported_os(self, mock_system):
        response = update_system_packages()
        self.assertEqual(response["error_code"], "UNSUPPORTED_OS")
        self.assertIsNone(response["output"])
        self.assertIn("Unsupported operating system: UnknownOS", response["error"])

    @patch('platform.system', return_value='Windows')
    @patch('subprocess.run', side_effect=Exception("Mocked exception"))
    def test_update_system_packages_unexpected_error(self, mock_run, mock_system):
        response = update_system_packages()
        mock_run.assert_called_with('winget upgrade --all', shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "FAILED_TO_UPDATE_PACKAGES")
        self.assertIsNone(response["output"])
        self.assertIn("Failed to update system packages: Mocked exception", response["error"])

if __name__ == "__main__":
    unittest.main()


# tools/find_file/test.py

import pytest
from unittest.mock import patch
from .api import find_file

@pytest.mark.asyncio
async def test_find_file_success():
    pattern = "*.py"
    depth = 2
    case_sensitive = False
    include = ["/test/include"]
    exclude = ["/test/exclude"]
    expected_matched_files = ["/test/include/file1.py", "/test/include/subdir/file2.py"]

    with patch('os.getcwd', return_value="/test"), \
         patch('os.path.isfile', return_value=True), \
         patch('os.path.expanduser', side_effect=lambda x: x), \
         patch('os.walk') as mock_walk, \
         patch('fnmatch.fnmatch') as mock_fnmatch:
        
        mock_walk.return_value = [
            ("/test/include", ["subdir"], ["file1.py"]),
            ("/test/include/subdir", [], ["file2.py"]),
            ("/test/exclude", [], ["file3.py"])
        ]
        
        # Mock fnmatch to handle case sensitivity
        def fnmatch_side_effect(filename, pattern):
            if not case_sensitive:
                return fnmatch.fnmatchcase(filename.lower(), pattern.lower())
            return fnmatch.fnmatchcase(filename, pattern)
        
        mock_fnmatch.side_effect = fnmatch_side_effect
        
        result = find_file(pattern, depth, case_sensitive, include, exclude)
        assert result['output'] == expected_matched_files
        assert result['error'] is None
        assert result['error_code'] == "SUCCESS"

@pytest.mark.asyncio
async def test_find_file_no_matches():
    pattern = "*.md"
    depth = 1
    case_sensitive = True
    include = ["/test/include"]
    exclude = ["/test/exclude"]
    expected_error = "No files found matching the pattern."

    with patch('os.getcwd', return_value="/test"), \
         patch('os.path.isfile', return_value=True), \
         patch('os.path.expanduser', side_effect=lambda x: x), \
         patch('os.walk') as mock_walk, \
         patch('fnmatch.fnmatch') as mock_fnmatch:
        
        mock_walk.return_value = [
            ("/test/include", [], ["file1.py", "file2.py"]),
            ("/test/exclude", [], ["file3.py"])
        ]
        
        mock_fnmatch.return_value = False
        
        result = find_file(pattern, depth, case_sensitive, include, exclude)
        assert result['output'] is None
        assert result['error'] == expected_error
        assert result['error_code'] == "NO_MATCHES"

@pytest.mark.asyncio
async def test_find_file_too_many_results():
    pattern = "*.txt"
    depth = None
    case_sensitive = False
    include = ["/test/include"]
    exclude = ["/test/exclude"]
    # Generate 250 mock files
    expected_matched_files = [f"/test/include/file{i}.txt" for i in range(1, 201)]
    total_files = 250

    with patch('os.getcwd', return_value="/test"), \
         patch('os.path.isfile', return_value=True), \
         patch('os.path.expanduser', side_effect=lambda x: x), \
         patch('os.walk') as mock_walk, \
         patch('fnmatch.fnmatch') as mock_fnmatch:
        
        mock_walk.return_value = [
            ("/test/include", [], [f"file{i}.txt" for i in range(1, total_files + 1)])
        ]
        
        mock_fnmatch.return_value = True
        
        result = find_file(pattern, depth, case_sensitive, include, exclude)
        assert result['output'] == expected_matched_files
        assert result['error'] == f"Too many results found ({total_files}). Showing first 200 results."
        assert result['error_code'] == "TOO_MANY_RESULTS"

@pytest.mark.asyncio
async def test_find_file_missing_pattern():
    pattern = ""
    depth = 1
    case_sensitive = True
    include = ["/test/include"]
    exclude = ["/test/exclude"]
    expected_error = "Pattern must be provided."

    result = find_file(pattern, depth, case_sensitive, include, exclude)
    assert result['output'] is None
    assert result['error'] == expected_error
    assert result['error_code'] == "MISSING_PATTERN"

@pytest.mark.asyncio
async def test_find_file_exception():
    pattern = "*.py"
    depth = 1
    case_sensitive = False
    include = ["/test/include"]
    exclude = ["/test/exclude"]
    expected_error_code = "FIND_FILE_FAILED"
    expected_error_message = "Error finding files: Sample exception."

    with patch('os.walk', side_effect=Exception("Sample exception.")):
        result = find_file(pattern, depth, case_sensitive, include, exclude)
        assert result['output'] is None
        assert result['error'] == expected_error_message
        assert result['error_code'] == expected_error_code


"""
We can run the command like below: (specify your own config file path)
> aeiva-chat-terminal --config configs/agent_config.yaml
"""
import os
import sys
import signal
import asyncio
from pathlib import Path
import click
from aeiva.agent.agent import Agent
from aeiva.util.file_utils import from_json_or_yaml
from aeiva.command.command_utils import (
    get_package_root,
    get_log_dir,
    setup_logging,
    validate_neo4j_home,
    start_neo4j,
    stop_neo4j,
    handle_exit,
)

# Get default agent config file path
PACKAGE_ROOT = get_package_root()
DEFAULT_CONFIG_PATH = PACKAGE_ROOT / 'configs' / 'agent_config.yaml'

# Get default log file path
LOGS_DIR = get_log_dir()
LOGS_DIR.mkdir(parents=True, exist_ok=True)  # Ensure the log directory exists
DEFAULT_LOG_PATH = LOGS_DIR / 'aeiva-chat-terminal.log'


@click.command()
@click.option('--config', '-c', default=str(DEFAULT_CONFIG_PATH),
              help='Path to the configuration file (YAML or JSON).',
              type=click.Path(exists=True, dir_okay=False))
@click.option('--verbose', '-v', is_flag=True, help='Enable verbose logging.')
def run(config, verbose):
    """
    Starts the Aeiva chat terminal with the provided configuration.
    """
    # Setup logging
    logger = setup_logging(DEFAULT_LOG_PATH, verbose)
    
    click.echo(f"Loading configuration from {config}")
    config_path = Path(config)
    
    # Parse the configuration file with error handling
    try:
        config_data = from_json_or_yaml(config_path)
    except Exception as e:
        logger.error(f"Failed to parse configuration file: {e}")
        click.echo(f"Error: Failed to parse configuration file: {e}")
        sys.exit(1)
    
    # Retrieve NEO4J_HOME from environment variables
    neo4j_home = os.getenv('NEO4J_HOME')
    if not neo4j_home:
        logger.error("NEO4J_HOME is not set in the environment.")
        click.echo("Error: NEO4J_HOME is not set in the environment. Please set it in your shell configuration (e.g., .bashrc or .zshrc).")
        sys.exit(1)
    
    # Validate NEO4J_HOME path
    validate_neo4j_home(logger, neo4j_home)
    
    # Start Neo4j
    neo4j_process = start_neo4j(logger, neo4j_home)
    
    # Register signal handlers to ensure Neo4j stops gracefully
    signal.signal(signal.SIGINT, lambda s, f: handle_exit(s, f, neo4j_process))
    signal.signal(signal.SIGTERM, lambda s, f: handle_exit(s, f, neo4j_process))
    
    # Start the Agent
    try:
        agent = Agent(config_data)
        agent.setup()
        asyncio.run(agent.run())
    except KeyboardInterrupt:
        logger.info("Agent execution interrupted by user.")
        click.echo("\nAgent execution interrupted by user.")
    except Exception as e:
        logger.error(f"An error occurred during agent execution: {e}")
        click.echo(f"An error occurred during agent execution: {e}")
    finally:
        # # Perform any necessary cleanup
        # try:
        #     agent.cognition_components['memory'].delete_all()
        #     logger.info("All memory units deleted during cleanup.")
        # except NotImplementedError as nie:
        #     logger.warning(f"Delete All feature not implemented: {nie}")
        # except Exception as e:
        #     logger.error(f"Error during cleanup: {e}")
        #     click.echo("Failed to delete all memory units.")
        
        # Stop Neo4j
        stop_neo4j(logger, neo4j_process)
        logger.info("Cleanup completed.")

if __name__ == "__main__":
    run()


# navigate_to_webpage/test.py

import pytest
from unittest.mock import patch, MagicMock
from navigate_to_webpage.api import navigate_to_webpage
from selenium.common.exceptions import WebDriverException

@pytest.mark.parametrize("request, mock_get_side_effect, expected", [
    # Successful navigation
    (
        {
            "url": "http://example.com",
            "timeout": 5000
        },
        None,  # No exception
        {
            "output": {"message": "Navigated to the specified webpage."},
            "error": None,
            "error_code": "SUCCESS"
        }
    ),
    # Missing required parameter
    (
        {
            "timeout": 5000
        },
        None,
        {
            "output": None,
            "error": "Missing required parameter: url",
            "error_code": "VALIDATION_ERROR"
        }
    ),
    # WebDriver exception during navigation
    (
        {
            "url": "http://example.com",
            "timeout": 5000
        },
        WebDriverException("Session not created"),
        {
            "output": None,
            "error": "WebDriver Error: Session not created",
            "error_code": "WEBDRIVER_ERROR"
        }
    ),
    # Unexpected exception
    (
        {
            "url": "http://example.com",
            "timeout": 5000
        },
        Exception("Unexpected error"),
        {
            "output": None,
            "error": "Unexpected Error: Unexpected error",
            "error_code": "UNEXPECTED_ERROR"
        }
    ),
])
def test_navigate_to_webpage(request, mock_get_side_effect, expected):
    with patch('navigate_to_webpage.api.webdriver.Chrome') as mock_webdriver:
        mock_driver = MagicMock()
        mock_webdriver.return_value = mock_driver

        if mock_get_side_effect:
            mock_driver.get.side_effect = mock_get_side_effect
        else:
            mock_driver.get.return_value = None

        # Mock driver.quit
        mock_driver.quit.return_value = None

        result = navigate_to_webpage(request)

        assert result == expected


from functools import partial

from transformers import LlamaTokenizer

from aeiva.common.constants import IGNORE_ID
from aeiva.config import OmniConfig
from aeiva.util.json_utils import dump_json
from aeiva.util.file_utils import ensure_dir
from aeiva.util.token_utils import get_tokenizer
from aeiva.operator.dataitem_ops import sample_frames_from_video, extract_audio_from_video, tokenize_and_label_text_for_instruction_tuning
from aeiva.operator.dataset_ops import build_dataset, merge_datasets, sample_dataset, filter_dataset_by_keys


if __name__ == "__main__":
    # load config
    config_path = "/Users/bangliu/Desktop/ChatSCI/Aeiva/configs/train_macaw.yaml"
    OmniConfig.create_omni_config()
    CONFIG = OmniConfig.from_yaml(config_path)
    print("config: ", CONFIG)
    print("==================== Step 0: loading config done")

    # shared by all datasets
    # tokenizer_name_or_path = 'yahma/llama-7b-hf'  #!!! currently different with the author. Maybe we can just load from the downloaded tokenizer.
    # TOKENIZER = get_tokenizer(tokenizer_name_or_path, add_special_tokens=True, special_tokens_dict=TOKENIZER_SPECIAL_TOKENS)
    TOKENIZER =  get_tokenizer("/Users/bangliu/Desktop/ChatSCI/Aeiva/pretrained_models/macaw/", tokenizer_cls=LlamaTokenizer)

    DATA_PROCESS_PIPELINE = [
        partial(tokenize_and_label_text_for_instruction_tuning, tokenizer=TOKENIZER, max_length=CONFIG.max_seq_len_for_preprocess, ignore_id=IGNORE_ID),
        partial(sample_frames_from_video, num_frames=CONFIG.num_frames_to_sample, video_dir=CONFIG.video_dir, frame_dir=CONFIG.frame_dir),  #!!! get frame indices
        partial(extract_audio_from_video, video_dir=CONFIG.video_dir, audio_dir=CONFIG.audio_dir)  # !!! get audio name
    ]

    # process each single dataset
    processed_datasets = {}

    # process avsd
    dataset_name = "avsd"
    input_filepaths_dict = {
        "avsd_dataset_path": "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/avsd/avsd_train.json"
    }
    output_dir = "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/avsd/output/"
    processed_data = build_dataset(dataset_name, input_filepaths_dict, DATA_PROCESS_PIPELINE, output_dir, CONFIG.num_samples_per_dataset)
    processed_datasets["avsd"] = processed_data

    # process alpaca
    dataset_name = "alpaca"
    input_filepaths_dict = {
        "alpaca_dataset_path": "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/alpaca/alpaca_data.json"
    }
    output_dir = "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/alpaca/output/"
    processed_data = build_dataset(dataset_name, input_filepaths_dict, DATA_PROCESS_PIPELINE, output_dir, CONFIG.num_samples_per_dataset)
    processed_datasets["alpaca"] = processed_data

    # process macaw_coco
    dataset_name = "macaw_coco"
    input_filepaths_dict = {
        "macaw_coco_dataset_path": "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/macaw/generated_examples_coco.json",
    }
    output_dir = "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/macaw/output/"
    processed_data = build_dataset(dataset_name, input_filepaths_dict, DATA_PROCESS_PIPELINE, output_dir, CONFIG.num_samples_per_dataset)
    processed_datasets["macaw_coco"] = processed_data

    # process macaw_avsd
    dataset_name = "macaw_avsd"
    input_filepaths_dict = {
        "macaw_avsd_dataset_path": "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/macaw/generated_examples_avsd.json",
    }
    output_dir = "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/macaw/output/"
    processed_data = build_dataset(dataset_name, input_filepaths_dict, DATA_PROCESS_PIPELINE, output_dir, CONFIG.num_samples_per_dataset)
    processed_datasets["macaw_avsd"] = processed_data

    # process vqa
    dataset_name = "vqa"
    input_filepaths_dict = {
        "vqa_annotations_path": "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/vqa/v2_mscoco_train2014_annotations.json",
        "vqa_questions_path": "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/vqa/v2_OpenEnded_mscoco_train2014_questions.json",
        }
    output_dir = "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/vqa/output/"
    processed_data = build_dataset(dataset_name, input_filepaths_dict, DATA_PROCESS_PIPELINE, output_dir, CONFIG.num_samples_per_dataset)
    processed_datasets["vqa"] = processed_data

    print("==================== Step 1: format and process datasets done")

    # merge and sample datasets
    keys_to_preserve = ['text', 'text_token_ids', 'attention_mask', 'labels', 'image', 'audio', 'video', 'sampled_video_frame_indices']  #!!!!!

    # merge 1
    datasets_to_merge = ["avsd", "alpaca", "vqa"]
    output_path = "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/merge/avsd_alpaca_vqa.json"
    merged_datasets = [processed_datasets[dataset_name] for dataset_name in datasets_to_merge]
    merged_datasets = merge_datasets(merged_datasets)
    merged_datasets = sample_dataset(merged_datasets, CONFIG.num_samples_per_merged_dataset)
    merged_datasets = filter_dataset_by_keys(merged_datasets, keys_to_preserve)
    ensure_dir(output_path)
    dump_json(merged_datasets, output_path)

    # merge 2
    datasets_to_merge = ["macaw_coco", "macaw_avsd"]
    output_path = "/Users/bangliu/Desktop/ChatSCI/Aeiva/datasets/merge/macaw_coco_avsd.json"
    merged_datasets = [processed_datasets[dataset_name] for dataset_name in datasets_to_merge]
    merged_datasets = merge_datasets(merged_datasets)
    merged_datasets = sample_dataset(merged_datasets, CONFIG.num_samples_per_merged_dataset)
    merged_datasets = filter_dataset_by_keys(merged_datasets, keys_to_preserve)
    ensure_dir(output_path)
    dump_json(merged_datasets, output_path)

    print("==================== Step 2: merge and sample datasets done")



# tools/test_operation/test.py

import pytest
from unittest.mock import patch
from .api import test_operation

@pytest.mark.asyncio
async def test_test_operation_success():
    a = 2
    b = 3
    expected_output = 2 + 3 + (2 * 3) + 100  # 2 + 3 + 6 + 100 = 111
    
    result = test_operation(a, b)
    assert result['output'] == expected_output
    assert result['error'] is None
    assert result['error_code'] == "SUCCESS"

@pytest.mark.asyncio
async def test_test_operation_failed():
    a = 2
    b = 3
    expected_error_code = "TEST_OPERATION_FAILED"
    expected_error_message = "Error performing test operation: Sample exception."
    
    with patch('tools.test_operation.api.test_operation', side_effect=Exception("Sample exception.")):
        result = test_operation(a, b)
        assert result['output'] is None
        assert result['error'] == expected_error_message
        assert result['error_code'] == expected_error_code


# toolkit/math_toolkit.py

from aeiva.tool.toolkit.toolkit import Toolkit

class MathToolkit(Toolkit):
    """
    A toolkit for mathematical operations.
    """

    def __init__(self, config=None):
        super().__init__(
            name="MathToolkit",
            tool_names=["calculator"],
            config=config
        )


from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional


class RelationalDatabase(ABC):
    """
    Abstract base class for relational database operations.
    """

    @abstractmethod
    def insert_record(self, table: str, record: Dict[str, Any]) -> Any:
        """
        Inserts a record into a table.

        Args:
            table (str): The name of the table.
            record (Dict[str, Any]): A dictionary representing the record to insert.

        Returns:
            Any: The primary key of the inserted record.

        Raises:
            StorageError: If there is an issue inserting the record.
        """
        pass

    @abstractmethod
    def get_record(self, table: str, primary_key: Any) -> Dict[str, Any]:
        """
        Retrieves a record by its primary key.

        Args:
            table (str): The name of the table.
            primary_key (Any): The primary key of the record.

        Returns:
            Dict[str, Any]: The retrieved record.

        Raises:
            RecordNotFoundError: If the record does not exist.
            StorageError: If there is an issue retrieving the record.
        """
        pass

    @abstractmethod
    def update_record(self, table: str, primary_key: Any, updates: Dict[str, Any]) -> None:
        """
        Updates a record in a table.

        Args:
            table (str): The name of the table.
            primary_key (Any): The primary key of the record.
            updates (Dict[str, Any]): A dictionary of fields to update.

        Raises:
            RecordNotFoundError: If the record does not exist.
            StorageError: If there is an issue updating the record.
        """
        pass

    @abstractmethod
    def delete_record(self, table: str, primary_key: Any) -> None:
        """
        Deletes a record from a table.

        Args:
            table (str): The name of the table.
            primary_key (Any): The primary key of the record.

        Raises:
            RecordNotFoundError: If the record does not exist.
            StorageError: If there is an issue deleting the record.
        """
        pass

    @abstractmethod
    def query_records(self, table: str, conditions: Optional[Dict[str, Any]] = None, limit: Optional[int] = None, offset: Optional[int] = None) -> List[Dict[str, Any]]:
        """
        Queries records from a table based on conditions.

        Args:
            table (str): The name of the table.
            conditions (Optional[Dict[str, Any]]): Conditions to filter records.
            limit (Optional[int]): Maximum number of records to return.
            offset (Optional[int]): Number of records to skip.

        Returns:
            List[Dict[str, Any]]: A list of records matching the query.

        Raises:
            StorageError: If there is an issue querying records.
        """
        pass

    @abstractmethod
    def execute_sql(self, query: str, parameters: Optional[Dict[str, Any]] = None) -> Any:
        """
        Executes a raw SQL query.

        Args:
            query (str): The SQL query string.
            parameters (Optional[Dict[str, Any]]): Parameters for parameterized queries.

        Returns:
            Any: The result of the query.

        Raises:
            StorageError: If there is an issue executing the query.
        """
        pass

    @abstractmethod
    def begin_transaction(self) -> None:
        """
        Begins a transaction.
        """
        pass

    @abstractmethod
    def commit_transaction(self) -> None:
        """
        Commits the current transaction.
        """
        pass

    @abstractmethod
    def rollback_transaction(self) -> None:
        """
        Rolls back the current transaction.
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Closes the database connection and releases resources.
        """
        pass


from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

class InputInterpreter(ABC):
    """
    Abstract base class for the Input Interpreter, responsible for processing incoming stimuli
    and determining how to format and forward them to the cognition system (LLM Brain).
    """

    def __init__(self, config: Optional[Dict[str, Any]] = None):
        self.config = config or {}

    @abstractmethod
    def gate(self, raw_input: Any) -> Any:
        """
        Lightweight gate that determines whether the input can be forwarded directly to the brain
        or if further processing is needed.
        """
        pass

    @abstractmethod
    def interpret(self, raw_input: Any) -> Dict[str, Any]:
        """
        Heavy processing that aggregates and formats the raw input into structured stimuli/prompt
        that the brain can better handle.
        """
        pass


# File: cognition/emotion_category.py

from dataclasses import dataclass, field
from typing import Dict, Any

@dataclass
class CategoryEmotionState:
    """
    Represents the emotional state in a Category-Based Model with extensive categories.
    
    Attributes:
        emotion_label (str): The current emotion category.
        intensity (float): The intensity of the current emotion (range: 0.0 to 1.0).
    """
    emotion_label: str = "neutral"
    intensity: float = 0.0  # Optional: Represents the strength of the emotion

    def to_dict(self) -> Dict[str, Any]:
        return {
            'emotion_label': self.emotion_label,
            'intensity': self.intensity
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        return CategoryEmotionState(
            emotion_label=data.get('emotion_label', 'neutral'),
            intensity=data.get('intensity', 0.0)
        )

# File: cognition/emotion_category.py

from typing import Dict, Any, Set
from aeiva.cognition.emotion.emotion import Emotion, ConfigurationError, UpdateError, RegulationError
import asyncio
import json
import logging

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class CategoryEmotion(Emotion[CategoryEmotionState]):
    """
    Concrete implementation of the Emotion base class using an extensive Category-Based Model.
    """
    
    # Define the extensive set of emotion categories
    EMOTION_CATEGORIES: Set[str] = {
        # Positive Emotions
        'admiration', 'amusement', 'approval', 'caring', 'desire', 'excitement',
        'gratitude', 'joy', 'love', 'optimism', 'pride', 'relief',
        # Negative Emotions
        'anger', 'annoyance', 'disappointment', 'disapproval', 'disgust',
        'embarrassment', 'fear', 'grief', 'nervousness', 'remorse', 'sadness',
        # Ambiguous Emotions
        'confusion', 'curiosity', 'realization', 'surprise', 'awe',
        # Additional Emotions
        'frustration', 'hope', 'guilt', 'contentment', 'boredom',
        'loneliness', 'elation'
    }
    
    def init_state(self) -> CategoryEmotionState:
        """
        Initialize the emotional state with a default emotion.
        
        Returns:
            CategoryEmotionState: Initialized state with 'neutral' emotion.
        """
        default_emotion = self.config.get('default_emotion', 'neutral')
        default_intensity = self.config.get('default_intensity', 0.0)
        if default_emotion not in self.EMOTION_CATEGORIES and default_emotion != 'neutral':
            raise ConfigurationError(f"Invalid default emotion: {default_emotion}")
        return CategoryEmotionState(emotion_label=default_emotion, intensity=default_intensity)
    
    async def setup(self) -> None:
        """
        Set up the Category Emotion system.
        
        Raises:
            ConfigurationError: If required configuration parameters are missing or invalid.
        """
        required_keys = ['allowed_emotions', 'regulation_strategies']
        for key in required_keys:
            if key not in self.config:
                raise ConfigurationError(f"Missing configuration parameter: {key}")
        
        # Initialize allowed emotions
        self.allowed_emotions = set(self.config['allowed_emotions'])
        if not self.allowed_emotions.issubset(self.EMOTION_CATEGORIES):
            invalid_emotions = self.allowed_emotions - self.EMOTION_CATEGORIES
            raise ConfigurationError(f"Invalid emotions in allowed_emotions: {invalid_emotions}")
        
        # Initialize regulation strategies
        self.regulation_strategies = set(self.config['regulation_strategies'])
        valid_strategies = {'neutralize', 'increase', 'decrease', 'toggle'}
        if not self.regulation_strategies.issubset(valid_strategies):
            invalid_strategies = self.regulation_strategies - valid_strategies
            raise ConfigurationError(f"Invalid regulation strategies: {invalid_strategies}")
        
        logging.info("CategoryEmotion setup complete.")
    
    async def update(self, input_data: Dict[str, Any]) -> None:
        """
        Update the emotional state based on input stimuli.
        
        Args:
            input_data (Dict[str, Any]): Stimuli data affecting emotion.
                Expected key: 'emotion_label' (str), 'intensity' (float, optional)
        
        Raises:
            UpdateError: If updating the emotional state fails.
        """
        try:
            new_emotion = input_data.get('emotion_label')
            if not new_emotion:
                raise ValueError("Missing 'emotion_label' in input_data.")
            if new_emotion not in self.allowed_emotions:
                raise ValueError(f"Emotion '{new_emotion}' is not allowed in this model.")
            
            new_intensity = input_data.get('intensity', 1.0)  # Default intensity
            
            if not (0.0 <= new_intensity <= 1.0):
                raise ValueError("Intensity must be between 0.0 and 1.0.")
            
            # Update the state
            self.state.emotion_label = new_emotion
            self.state.intensity = new_intensity
            
            logging.debug(f"Updated emotional state: {self.state.to_dict()}")
        except Exception as e:
            self.handle_error(e)
            raise UpdateError("Failed to update emotional state.") from e
    
    def regulate(self, strategy: str, emotion_label: str = None, value: float = None) -> None:
        """
        Regulate the emotional state using a specified strategy.
        
        Args:
            strategy (str): The regulation strategy to apply ('neutralize', 'increase', 'decrease', 'toggle').
            emotion_label (str, optional): The emotion to regulate (required for 'increase' and 'decrease').
            value (float, optional): The value to set or adjust the intensity to (required for 'set' operations).
        
        Raises:
            RegulationError: If the regulation strategy is invalid or fails.
        """
        try:
            if strategy not in self.regulation_strategies:
                raise ValueError(f"Regulation strategy '{strategy}' is not supported.")
            
            if strategy == 'neutralize':
                self.state.emotion_label = 'neutral'
                self.state.intensity = 0.0
                logging.info("Applied regulation strategy: neutralize")
            elif strategy == 'increase':
                if not emotion_label:
                    raise ValueError("Missing 'emotion_label' for 'increase' strategy.")
                if emotion_label not in self.allowed_emotions:
                    raise ValueError(f"Emotion '{emotion_label}' is not allowed in this model.")
                if value is None:
                    raise ValueError("Missing 'value' for 'increase' strategy.")
                if not (0.0 <= value <= 1.0):
                    raise ValueError("Value must be between 0.0 and 1.0.")
                if self.state.emotion_label == emotion_label:
                    self.state.intensity = min(self.state.intensity + value, 1.0)
                else:
                    self.state.emotion_label = emotion_label
                    self.state.intensity = value
                logging.info(f"Applied regulation strategy: increase '{emotion_label}' by {value}")
            elif strategy == 'decrease':
                if not emotion_label:
                    raise ValueError("Missing 'emotion_label' for 'decrease' strategy.")
                if emotion_label not in self.allowed_emotions:
                    raise ValueError(f"Emotion '{emotion_label}' is not allowed in this model.")
                if value is None:
                    raise ValueError("Missing 'value' for 'decrease' strategy.")
                if not (0.0 <= value <= 1.0):
                    raise ValueError("Value must be between 0.0 and 1.0.")
                if self.state.emotion_label == emotion_label:
                    self.state.intensity = max(self.state.intensity - value, 0.0)
                    if self.state.intensity == 0.0:
                        self.state.emotion_label = 'neutral'
                logging.info(f"Applied regulation strategy: decrease '{emotion_label}' by {value}")
            elif strategy == 'toggle':
                if self.state.emotion_label != 'neutral':
                    self.state.emotion_label = 'neutral'
                    self.state.intensity = 0.0
                    logging.info("Applied regulation strategy: toggle to 'neutral'")
                else:
                    raise ValueError("Cannot toggle 'neutral' to another emotion without specifying.")
            else:
                raise ValueError(f"Unknown regulation strategy: {strategy}")
        except Exception as e:
            self.handle_error(e)
            raise RegulationError(f"Failed to apply regulation strategy: {strategy}") from e
    
    def express(self) -> str:
        """
        Generate a textual expression of the current emotional state.
        
        Returns:
            str: A string describing the current emotion.
        """
        if self.state.emotion_label == 'neutral' or self.state.intensity == 0.0:
            return "I'm feeling neutral."
        else:
            # Optionally, adjust the expression based on intensity
            intensity_description = ""
            if self.state.intensity >= 0.75:
                intensity_description = " intensely"
            elif self.state.intensity >= 0.5:
                intensity_description = ""
            else:
                intensity_description = " slightly"
            
            return f"I feel{intensity_description} {self.state.emotion_label}."
    
    def serialize(self) -> str:
        """
        Serialize the current emotional state to a JSON string.
        
        Returns:
            str: Serialized emotional state.
        """
        return json.dumps(self.state.to_dict())
    
    def deserialize(self, data: str) -> None:
        """
        Deserialize the emotional state from a JSON string.
        
        Args:
            data (str): Serialized emotional state.
        """
        try:
            state_dict = json.loads(data)
            self.state = CategoryEmotionState.from_dict(state_dict)
            logging.info("Deserialized emotional state.")
        except Exception as e:
            self.handle_error(e)
            raise UpdateError("Failed to deserialize emotional state.") from e

# File: main_category_emotion.py

import asyncio
from aeiva.cognition.emotion.emotion_category import CategoryEmotion

async def main():
    # Configuration for CategoryEmotion
    config = {
        'allowed_emotions': [
            # Positive Emotions
            'admiration', 'amusement', 'approval', 'caring', 'desire', 'excitement',
            'gratitude', 'joy', 'love', 'optimism', 'pride', 'relief',
            # Negative Emotions
            'anger', 'annoyance', 'disappointment', 'disapproval', 'disgust',
            'embarrassment', 'fear', 'grief', 'nervousness', 'remorse', 'sadness',
            # Ambiguous Emotions
            'confusion', 'curiosity', 'realization', 'surprise', 'awe',
            # Additional Emotions
            'frustration', 'hope', 'guilt', 'contentment', 'boredom',
            'loneliness', 'elation'
        ],
        'regulation_strategies': ['neutralize', 'increase', 'decrease', 'toggle'],
        'default_emotions': {
            'emotion_label': 'neutral',
            'intensity': 0.0
        }
    }

    # Instantiate and set up the emotion system
    emotion_system = CategoryEmotion(config)
    await emotion_system.setup()

    # Express initial state
    print(emotion_system.express())  # Output: I'm feeling neutral.

    # Update with stimuli: high joy
    await emotion_system.update({
        'emotion_label': 'joy',
        'intensity': 0.8
    })
    print(emotion_system.express())  # Output: I feel intensely joy.

    # Update with stimuli: moderate curiosity
    await emotion_system.update({
        'emotion_label': 'curiosity',
        'intensity': 0.5
    })
    print(emotion_system.express())  # Output: I feel intensely curiosity.

    # Apply regulation: decrease intensity of 'curiosity' by 0.3
    emotion_system.regulate('decrease', emotion_label='curiosity', value=0.3)
    print(emotion_system.express())  # Output: I feel slightly curiosity.

    # Apply regulation: neutralize emotions
    emotion_system.regulate('neutralize')
    print(emotion_system.express())  # Output: I'm feeling neutral.

    # Serialize the emotional state
    serialized_state = emotion_system.serialize()
    print(f"Serialized State: {serialized_state}")

    # Deserialize into a new emotion system instance
    new_emotion_system = CategoryEmotion(config)
    await new_emotion_system.setup()
    new_emotion_system.deserialize(serialized_state)
    print(new_emotion_system.express())  # Output: I'm feeling neutral.

    # Example of toggling emotion
    await emotion_system.update({
        'emotion_label': 'hope',
        'intensity': 0.6
    })
    print(emotion_system.express())  # Output: I feel hope.
    
    emotion_system.regulate('toggle')
    print(emotion_system.express())  # Output: I'm feeling neutral.

    # Example of increasing an emotion's intensity
    emotion_system.regulate('increase', emotion_label='hope', value=0.3)
    print(emotion_system.express())  # Output: I feel intensely hope.

if __name__ == '__main__':
    asyncio.run(main())



# toolkit/system_toolkit/get_disk_usage/model.py

from enum import Enum
from pydantic import BaseModel, Field

class GetDiskUsageParams(BaseModel):
    path: str = Field(
        "/",
        description="The path to check disk usage for."
    )

class GetDiskUsageResult(BaseModel):
    output: dict = Field(
        ..., 
        description="Disk usage information."
    )
    error: str = Field(
        None, 
        description="Error message, if any."
    )
    error_code: str = Field(
        ..., 
        description="Error code representing the result state."
    )

class GetDiskUsageErrorCode(str, Enum):
    SUCCESS = "SUCCESS"
    PATH_NOT_FOUND = "PATH_NOT_FOUND"
    FAILED_TO_GET_DISK_USAGE = "FAILED_TO_GET_DISK_USAGE"
    UNEXPECTED_ERROR = "UNEXPECTED_ERROR"


# toolkit/system_toolkit/view_system_logs/model.py

from enum import Enum
from pydantic import BaseModel, Field

class ViewSystemLogsParams(BaseModel):
    log_type: str = Field(
        "system",
        description="Type of logs to retrieve ('system' or 'application'). Defaults to 'system'."
    )

class ViewSystemLogsResult(BaseModel):
    output: dict = Field(
        ..., 
        description="Retrieved system logs."
    )
    error: str = Field(
        None, 
        description="Error message, if any."
    )
    error_code: str = Field(
        ..., 
        description="Error code representing the result state."
    )

class ViewSystemLogsErrorCode(str, Enum):
    SUCCESS = "SUCCESS"
    FAILED_TO_VIEW_LOGS = "FAILED_TO_VIEW_LOGS"
    UNSUPPORTED_OS = "UNSUPPORTED_OS"
    UNSUPPORTED_LOG_TYPE = "UNSUPPORTED_LOG_TYPE"
    UNEXPECTED_ERROR = "UNEXPECTED_ERROR"


# model.py

from pydantic import BaseModel, Field
from typing import Optional, Dict, Any

class EditFileErrorCode:
    FILE_NOT_FOUND = "FILE_NOT_FOUND"
    INVALID_LINE_RANGE = "INVALID_LINE_RANGE"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    OS_ERROR = "OS_ERROR"
    INVALID_PARAMETERS = "INVALID_PARAMETERS"
    UNEXPECTED_ERROR = "UNEXPECTED_ERROR"
    SUCCESS = "SUCCESS"

class EditFileParams(BaseModel):
    file_path: Optional[str] = Field(None, description="The path to the file that will be edited. If not provided, edits the currently open file.")
    text: str = Field(..., description="The text that will replace the specified line range in the file.")
    start_line: int = Field(..., description="The line number at which the file edit will start (inclusive).")
    end_line: Optional[int] = Field(None, description="The line number at which the file edit will end (inclusive). If not provided, appends the text.")

class EditFileResult(BaseModel):
    output: Optional[str] = Field(None, description="Message indicating the result of the edit action.")
    error: Optional[str] = Field(None, description="Error message if any.")
    error_code: Optional[str] = Field(None, description="Error code indicating the result status.")


# toolkit/system_toolkit/delete_user/test.py

import unittest
from unittest.mock import patch, MagicMock
from toolkit.system_toolkit.delete_user.api import delete_user

class TestDeleteUser(unittest.TestCase):
    @patch('platform.system', return_value='Linux')
    @patch('subprocess.run')
    def test_delete_user_success_linux(self, mock_run, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="userdel executed successfully.", stderr="")
        response = delete_user("testuser")
        mock_run.assert_called_with(['sudo', 'userdel', '-r', 'testuser'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"], "User 'testuser' deleted successfully.")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Darwin')
    @patch('subprocess.run')
    def test_delete_user_success_mac(self, mock_run, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="userdel executed successfully.", stderr="")
        response = delete_user("testuser")
        mock_run.assert_called_with(['sudo', 'userdel', '-r', 'testuser'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"], "User 'testuser' deleted successfully.")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Windows')
    @patch('subprocess.run')
    def test_delete_user_success_windows(self, mock_run, mock_system):
        mock_run.return_value = MagicMock(returncode=0, stdout="The command completed successfully.", stderr="")
        response = delete_user("testuser")
        mock_run.assert_called_with(['net', 'user', 'testuser', '/delete'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "SUCCESS")
        self.assertEqual(response["output"], "User 'testuser' deleted successfully.")
        self.assertIsNone(response["error"])

    @patch('platform.system', return_value='Linux')
    @patch('subprocess.run', side_effect=subprocess.CalledProcessError(1, ['userdel'], stderr="Error deleting user."))
    def test_delete_user_failed_linux(self, mock_run, mock_system):
        response = delete_user("testuser")
        mock_run.assert_called_with(['sudo', 'userdel', '-r', 'testuser'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "FAILED_TO_DELETE_USER")
        self.assertIsNone(response["output"])
        self.assertIn("Failed to delete user 'testuser': Error deleting user.", response["error"])

    @patch('platform.system', return_value='UnknownOS')
    def test_delete_user_unsupported_os(self, mock_system):
        response = delete_user("testuser")
        self.assertEqual(response["error_code"], "UNSUPPORTED_OS")
        self.assertIsNone(response["output"])
        self.assertIn("Unsupported operating system: UnknownOS", response["error"])

    @patch('platform.system', return_value='Linux')
    @patch('subprocess.run', side_effect=Exception("Mocked exception"))
    def test_delete_user_unexpected_error(self, mock_run, mock_system):
        response = delete_user("testuser")
        mock_run.assert_called_with(['sudo', 'userdel', '-r', 'testuser'], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        self.assertEqual(response["error_code"], "FAILED_TO_DELETE_USER")
        self.assertIsNone(response["output"])
        self.assertIn("Failed to delete user 'testuser': Mocked exception", response["error"])

if __name__ == "__main__":
    unittest.main()